#!/usr/bin/env python
# vim:fileencoding=utf-8:noet

'''Gradients generator
'''

from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import json
import argparse

from itertools import groupby

from colormath.color_objects import sRGBColor, LabColor
from colormath.color_conversions import convert_color
from colormath.color_diff import delta_e_cie2000

from powerline.colorscheme import cterm_to_hex


def num2(s):
	try:
		return (True, [int(v) for v in s.partition(' ')[::2]])
	except TypeError:
		return (False, [float(v) for v in s.partition(' ')[::2]])


def rgbint_to_lab(rgbint):
	rgb = sRGBColor(
		(rgbint >> 16) & 0xFF, (rgbint >> 8) & 0xFF, rgbint & 0xFF,
		is_upscaled=True
	)
	return convert_color(rgb, LabColor)


cterm_to_lab = tuple((rgbint_to_lab(v) for v in cterm_to_hex))


def color(s):
	if len(s) <= 3:
		return cterm_to_lab[int(s)]
	else:
		return rgbint_to_lab(int(s, 16))


def nums(s):
	return [int(i) for i in s.split()]


def linear_gradient(start_value, stop_value, start_offset, stop_offset, offset):
	return start_value + ((offset - start_offset) * (stop_value - start_value) / (stop_offset - start_offset))


def lab_gradient(slab, elab, soff, eoff, off):
	svals = slab.get_value_tuple()
	evals = elab.get_value_tuple()
	return LabColor(*[
		linear_gradient(start_value, end_value, soff, eoff, off)
		for start_value, end_value in zip(svals, evals)
	])


def generate_gradient_function(DATA):
	def gradient_function(y):
		initial_offset = 0
		for offset, start, end in DATA:
			if y <= offset:
				return lab_gradient(start, end, initial_offset, offset, y)
			initial_offset = offset
	return gradient_function


def get_upscaled_values(rgb):
	return [min(max(0, i), 255) for i in rgb.get_upscaled_value_tuple()]


def get_rgb(lab):
	rgb = convert_color(lab, sRGBColor)
	rgb = sRGBColor(*get_upscaled_values(rgb), is_upscaled=True)
	return rgb.get_rgb_hex()[1:]


def find_color(ulab, colors, ctrans):
	cur_distance = float('inf')
	cur_color = None
	i = 0
	for clab in colors:
		dist = delta_e_cie2000(ulab, clab)
		if dist < cur_distance:
			cur_distance = dist
			cur_color = (ctrans(i), clab)
		i += 1
	return cur_color


def print_color(color):
	if type(color) is int:
		colstr = '5;' + str(color)
	else:
		rgb = convert_color(color, sRGBColor)
		colstr = '2;' + ';'.join((str(i) for i in get_upscaled_values(rgb)))
	sys.stdout.write('\033[48;' + colstr + 'm ')


def print_colors(colors, num):
	for i in range(num):
		color = colors[int(round(i * (len(colors) - 1) / num))]
		print_color(color)
	sys.stdout.write('\033[0m\n')


def dec_scale_generator(num):
	j = 0
	r = ''
	while num:
		r += '\033[{0}m'.format(j % 2)
		for i in range(10):
			r += str(i)
			num -= 1
			if not num:
				break
		j += 1
	r += '\033[0m\n'
	return r


def compute_steps(gradient, weights):
	maxweight = len(gradient) - 1
	if weights:
		weight_sum = sum(weights)
		norm_weights = [100.0 * weight / weight_sum for weight in weights]
		steps = [0]
		for weight in norm_weights:
			steps.append(steps[-1] + weight)
		steps.pop(0)
		steps.pop(0)
	else:
		step = m / maxweight
		steps = [i * step for i in range(1, maxweight + 1)]
	return steps


palettes = {
	'16': (cterm_to_lab[:16], lambda c: c),
	'256': (cterm_to_lab, lambda c: c),
	None: (cterm_to_lab[16:], lambda c: c + 16),
}


def show_scale(rng, num_output):
	if not rng and num_output >= 32 and (num_output - 1) // 10 >= 4 and (num_output - 1) % 10 == 0:
		sys.stdout.write('0')
		sys.stdout.write(''.join(('%*u' % (num_output // 10, i) for i in range(10, 101, 10))))
		sys.stdout.write('\n')
	else:
		if rng:
			vmin, vmax = rng[1]
			isint = rng[0]
		else:
			isint = True
			vmin = 0
			vmax = 100
		s = ''
		lasts = ' ' + str(vmax)
		while len(s) + len(lasts) < num_output:
			curpc = len(s) + 1 if s else 0
			curval = vmin + curpc * (vmax - vmin) / num_output
			if isint:
				curval = int(round(curval))
			s += str(curval) + ' '
		sys.stdout.write(s[:-1] + lasts + '\n')
	sys.stdout.write(dec_scale_generator(num_output) + '\n')


if __name__ == '__main__':
	p = argparse.ArgumentParser(description=__doc__)
	p.add_argument('gradient', nargs='*', metavar='COLOR', type=color, help='List of colors (either indexes from 8-bit palette or 24-bit RGB in hexadecimal notation)')
	p.add_argument('-n', '--num_items', metavar='INT', type=int, help='Number of items in resulting list', default=101)
	p.add_argument('-N', '--num_output', metavar='INT', type=int, help='Number of characters in sample', default=101)
	p.add_argument('-r', '--range', metavar='V1 V2', type=num2, help='Use this range when outputting scale')
	p.add_argument('-s', '--show', action='store_true', help='If present output gradient sample')
	p.add_argument('-p', '--palette', choices=('16', '256'), help='Use this palette. Defaults to 240-color palette (256 colors without first 16)')
	p.add_argument('-w', '--weights', metavar='INT INT ...', type=nums, help='Adjust weights of colors. Number of weights must be equal to number of colors')
	p.add_argument('-C', '--omit-terminal', action='store_true', help='If present do not compute values for terminal')

	args = p.parse_args()

	m = args.num_items

	steps = compute_steps(args.gradient, args.weights)

	data = [
		(weight, args.gradient[i - 1], args.gradient[i])
		for weight, i in zip(steps, range(1, len(args.gradient)))
	]
	gr_func = generate_gradient_function(data)
	gradient = [gr_func(y) for y in range(0, m)]

	r = [get_rgb(lab) for lab in gradient]
	if not args.omit_terminal:
		r2 = [find_color(lab, *palettes[args.palette])[0] for lab in gradient]
		r3 = [i[0] for i in groupby(r2)]

	if not args.omit_terminal:
		print(json.dumps(r3) + ',')
		print(json.dumps(r2) + ',')
	print(json.dumps(r))

	if args.show:
		print_colors(args.gradient, args.num_output)
		if not args.omit_terminal:
			print_colors(r3, args.num_output)
			print_colors(r2, args.num_output)
		print_colors(gradient, args.num_output)

		show_scale(args.range, args.num_output)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.renderers.shell import ShellRenderer


ESCAPE_CHAR = '\001'


class KshPromptRenderer(ShellRenderer):
	'''Powerline bash prompt segment renderer.'''
	escape_hl_start = '\001'
	escape_hl_end = '\001'

	def render(self, *args, **kwargs):
		return '\001\r' + super(KshPromptRenderer, self).render(*args, **kwargs)


renderer = KshPromptRenderer



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import os

if sys.version_info < (2, 7):
	from unittest2 import TestCase as _TestCase  # NOQA
	from unittest2 import main as _main  # NOQA
	from unittest2.case import SkipTest  # NOQA
else:
	from unittest import TestCase as _TestCase  # NOQA
	from unittest import main as _main  # NOQA
	from unittest.case import SkipTest  # NOQA

from tests.modules.lib import PowerlineSingleTest


class PowerlineDummyTest(object):
	def __enter__(self):
		return self

	def __exit__(self, *args):
		pass

	def fail(self, *args, **kwargs):
		pass

	def exception(self, *args, **kwargs):
		pass


class PowerlineTestSuite(object):
	def __init__(self, name):
		self.name = name
		self.suite = ''

	def __enter__(self):
		self.saved_current_suite = os.environ['POWERLINE_CURRENT_SUITE']
		os.environ['POWERLINE_CURRENT_SUITE'] = (
			self.saved_current_suite + '/' + self.name)
		self.suite = self.saved_current_suite + '/' + self.name
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		if exc_type is not None:
			self.exception(
				'suite_noexcept',
				'Exception while running test suite: {0!r}'.format(exc_value),
			)
		os.environ['POWERLINE_CURRENT_SUITE'] = self.saved_current_suite

	def record_test_failure(self, fail_char, test_name, message, allow_failure=False):
		if allow_failure:
			fail_char = 'A' + fail_char
		full_msg = '{fail_char} {suite}|{test_name} :: {message}'.format(
			fail_char=fail_char,
			suite=self.suite,
			test_name=test_name,
			message=message,
		)
		with open(os.environ['FAILURES_FILE'], 'a') as ffd:
			ffd.write(full_msg + '\n')
		return False

	def exception(self, test_name, message, allow_failure=False):
		return self.record_test_failure('E', test_name, message, allow_failure)

	def fail(self, test_name, message, allow_failure=False):
		return self.record_test_failure('F', test_name, message, allow_failure)

	def test(self, name, attempts_left=0):
		if not attempts_left:
			return PowerlineSingleTest(self, name)
		else:
			return PowerlineDummyTest()

	def subsuite(self, name):
		return PowerlineTestSuite(name)


suite = None


def main(*args, **kwargs):
	global suite
	suite = PowerlineTestSuite(sys.argv[0])
	_main(*args, **kwargs)


class TestCase(_TestCase):
	def fail(self, msg=None):
		suite.fail(self.__class__.__name__,
		           msg or 'Test failed without message')
		super(TestCase, self).fail(*args, **kwargs)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os

from threading import Lock
from copy import deepcopy
from time import sleep
from functools import wraps

from powerline.renderer import Renderer
from powerline.lib.config import ConfigLoader
from powerline import Powerline, get_default_theme

from tests.modules.lib import Args, replace_attr


UT = get_default_theme(is_unicode=True)
AT = get_default_theme(is_unicode=False)


class TestHelpers(object):
	def __init__(self, config):
		self.config = config
		self.access_log = []
		self.access_lock = Lock()

	def loader_condition(self, path):
		return (path in self.config) and path

	def find_config_files(self, cfg_path, config_loader, loader_callback):
		if cfg_path.endswith('.json'):
			cfg_path = cfg_path[:-5]
		if cfg_path.startswith('/'):
			cfg_path = cfg_path.lstrip('/')
		with self.access_lock:
			self.access_log.append('check:' + cfg_path)
		if cfg_path in self.config:
			yield cfg_path
		else:
			if config_loader:
				config_loader.register_missing(self.loader_condition, loader_callback, cfg_path)
			raise IOError(('fcf:' if cfg_path.endswith('raise') else '') + cfg_path)

	def load_json_config(self, config_file_path, *args, **kwargs):
		if config_file_path.endswith('.json'):
			config_file_path = config_file_path[:-5]
		if config_file_path.startswith('/'):
			config_file_path = config_file_path.lstrip('/')
		with self.access_lock:
			self.access_log.append('load:' + config_file_path)
		try:
			return deepcopy(self.config[config_file_path])
		except KeyError:
			raise IOError(config_file_path)

	def pop_events(self):
		with self.access_lock:
			r = self.access_log[:]
			self.access_log = []
		return r


def log_call(func):
	@wraps(func)
	def ret(self, *args, **kwargs):
		self._calls.append((func.__name__, args, kwargs))
		return func(self, *args, **kwargs)
	return ret


class TestWatcher(object):
	events = set()
	lock = Lock()

	def __init__(self):
		self._calls = []

	@log_call
	def watch(self, file):
		pass

	@log_call
	def __call__(self, file):
		with self.lock:
			if file in self.events:
				self.events.remove(file)
				return True
		return False

	def _reset(self, files):
		with self.lock:
			self.events.clear()
			self.events.update(files)

	@log_call
	def unsubscribe(self):
		pass


class Logger(object):
	def __init__(self):
		self.messages = []
		self.lock = Lock()

	def _add_msg(self, attr, msg):
		with self.lock:
			self.messages.append(attr + ':' + msg)

	def _pop_msgs(self):
		with self.lock:
			r = self.messages
			self.messages = []
		return r

	def __getattr__(self, attr):
		return lambda *args, **kwargs: self._add_msg(attr, *args, **kwargs)


class SimpleRenderer(Renderer):
	def hlstyle(self, fg=None, bg=None, attrs=None):
		return '<{fg} {bg} {attrs}>'.format(fg=fg and fg[0], bg=bg and bg[0], attrs=attrs)


class EvenSimplerRenderer(Renderer):
	def hlstyle(self, fg=None, bg=None, attrs=None):
		return '{{{fg}{bg}{attrs}}}'.format(
			fg=fg and fg[0] or '-',
			bg=bg and bg[0] or '-',
			attrs=attrs if attrs else '',
		)


class TestPowerline(Powerline):
	_created = False

	def __init__(self, _helpers, **kwargs):
		super(TestPowerline, self).__init__(**kwargs)
		self._helpers = _helpers
		self.find_config_files = _helpers.find_config_files

	@staticmethod
	def get_local_themes(local_themes):
		return local_themes

	@staticmethod
	def get_config_paths():
		return ['']

	def _will_create_renderer(self):
		return self.cr_kwargs

	def _pop_events(self):
		return self._helpers.pop_events()


renderer = EvenSimplerRenderer


class TestConfigLoader(ConfigLoader):
	def __init__(self, _helpers, **kwargs):
		watcher = TestWatcher()
		super(TestConfigLoader, self).__init__(
			load=_helpers.load_json_config,
			watcher=watcher,
			watcher_type='test',
			**kwargs
		)


def get_powerline(config, **kwargs):
	helpers = TestHelpers(config)
	return get_powerline_raw(
		helpers,
		TestPowerline,
		_helpers=helpers,
		ext='test',
		renderer_module='tests.modules.lib.config_mock',
		logger=Logger(),
		**kwargs
	)


def select_renderer(simpler_renderer=False):
	global renderer
	renderer = EvenSimplerRenderer if simpler_renderer else SimpleRenderer


def get_powerline_raw(helpers, PowerlineClass, replace_gcp=False, **kwargs):
	if not isinstance(helpers, TestHelpers):
		helpers = TestHelpers(helpers)
	select_renderer(kwargs.pop('simpler_renderer', False))

	if replace_gcp:
		class PowerlineClass(PowerlineClass):
			@staticmethod
			def get_config_paths():
				return ['/']

	pl = PowerlineClass(
		config_loader=TestConfigLoader(
			_helpers=helpers,
			run_once=kwargs.get('run_once')
		),
		**kwargs
	)
	pl._watcher = pl.config_loader.watcher
	return pl


def swap_attributes(config, powerline_module):
	return replace_attr(powerline_module, 'os', Args(
		path=Args(
			isfile=lambda path: path.lstrip('/').replace('.json', '') in config,
			join=os.path.join,
			expanduser=lambda path: path,
			realpath=lambda path: path,
			dirname=os.path.dirname,
		),
		environ={},
	))


def add_watcher_events(p, *args, **kwargs):
	if isinstance(p._watcher, TestWatcher):
		p._watcher._reset(args)
	while not p._will_create_renderer():
		sleep(kwargs.get('interval', 0.1))
		if not kwargs.get('wait', True):
			return



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import os

from colormath.color_objects import sRGBColor, LabColor
from colormath.color_conversions import convert_color
from colormath.color_diff import delta_e_cie2000


def get_lab(name, rgb):
	rgb = sRGBColor(
		int(rgb[:2], 16), int(rgb[2:4], 16), int(rgb[4:6], 16),
		is_upscaled=True
	)
	lab = convert_color(rgb, LabColor)
	return name, lab


with open(os.path.join(os.path.dirname(__file__), 'colors.map'), 'r') as f:
	colors = [get_lab(*line.split('\t')) for line in f]


ulab = get_lab(None, sys.argv[1])[1]


def find_color(urgb, colors):
	cur_distance = 3 * (255 ** 2 + 1)
	cur_color = None
	for color, clab in colors:
		dist = delta_e_cie2000(ulab, clab)
		if dist < cur_distance:
			cur_distance = dist
			cur_color = (color, clab)
	return cur_color


cur_color = find_color(ulab, colors)


def lab_to_csi(lab):
	rgb = convert_color(lab, sRGBColor)
	colstr = ';2;' + ';'.join((str(i) for i in get_upscaled_values(rgb)))
	return colstr + 'm'


def get_upscaled_values(rgb):
	return [min(max(0, i), 255) for i in rgb.get_upscaled_value_tuple()]


def get_rgb(lab):
	rgb = convert_color(lab, sRGBColor)
	rgb = sRGBColor(*get_upscaled_values(rgb), is_upscaled=True)
	return rgb.get_rgb_hex()[1:]

print(get_rgb(ulab), ':', cur_color[0], ':', get_rgb(cur_color[1]))

col_1 = lab_to_csi(ulab)
col_2 = lab_to_csi(cur_color[1])
sys.stdout.write('\033[48' + col_1 + '\033[38' + col_2 + 'abc\033[0m <-- bg:urgb, fg:crgb\n')
sys.stdout.write('\033[48' + col_2 + '\033[38' + col_1 + 'abc\033[0m <-- bg:crgb, fg:urgb\n')



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from threading import Thread, Lock, Event
from types import MethodType

from powerline.lib.monotonic import monotonic
from powerline.segments import Segment


class MultiRunnedThread(object):
	daemon = True

	def __init__(self):
		self.thread = None

	def is_alive(self):
		return self.thread and self.thread.is_alive()

	def start(self):
		self.shutdown_event.clear()
		self.thread = Thread(target=self.run)
		self.thread.daemon = self.daemon
		self.thread.start()

	def join(self, *args, **kwargs):
		if self.thread:
			return self.thread.join(*args, **kwargs)
		return None


class ThreadedSegment(Segment, MultiRunnedThread):
	min_sleep_time = 0.1
	update_first = True
	interval = 1
	daemon = False

	argmethods = ('render', 'set_state')

	def __init__(self):
		super(ThreadedSegment, self).__init__()
		self.run_once = True
		self.crashed = False
		self.crashed_value = None
		self.update_value = None
		self.updated = False

	def __call__(self, pl, update_first=True, **kwargs):
		if self.run_once:
			self.pl = pl
			self.set_state(**kwargs)
			update_value = self.get_update_value(True)
		elif not self.is_alive():
			# Without this we will not have to wait long until receiving bug “I 
			# opened vim, but branch information is only shown after I move 
			# cursor”.
			#
			# If running once .update() is called in __call__.
			self.start()
			update_value = self.get_update_value(self.do_update_first)
		else:
			update_value = self.get_update_value(not self.updated)

		if self.crashed:
			return self.crashed_value

		return self.render(update_value, update_first=update_first, pl=pl, **kwargs)

	def set_update_value(self):
		try:
			self.update_value = self.update(self.update_value)
		except Exception as e:
			self.exception('Exception while updating: {0}', str(e))
			self.crashed = True
		except KeyboardInterrupt:
			self.warn('Caught keyboard interrupt while updating')
			self.crashed = True
		else:
			self.crashed = False
			self.updated = True

	def get_update_value(self, update=False):
		if update:
			self.set_update_value()
		return self.update_value

	def run(self):
		if self.do_update_first:
			start_time = monotonic()
			while True:
				self.shutdown_event.wait(max(self.interval - (monotonic() - start_time), self.min_sleep_time))
				if self.shutdown_event.is_set():
					break
				start_time = monotonic()
				self.set_update_value()
		else:
			while not self.shutdown_event.is_set():
				start_time = monotonic()
				self.set_update_value()
				self.shutdown_event.wait(max(self.interval - (monotonic() - start_time), self.min_sleep_time))

	def shutdown(self):
		self.shutdown_event.set()
		if self.daemon and self.is_alive():
			# Give the worker thread a chance to shutdown, but don’t block for 
			# too long
			self.join(0.01)

	def set_interval(self, interval=None):
		# Allowing “interval” keyword in configuration.
		# Note: Here **kwargs is needed to support foreign data, in subclasses 
		# it can be seen in a number of places in order to support 
		# .set_interval().
		interval = interval or getattr(self, 'interval')
		self.interval = interval

	def set_state(self, interval=None, update_first=True, shutdown_event=None, **kwargs):
		self.set_interval(interval)
		self.shutdown_event = shutdown_event or Event()
		self.do_update_first = update_first and self.update_first
		self.updated = self.updated or (not self.do_update_first)

	def startup(self, pl, **kwargs):
		self.run_once = False
		self.pl = pl
		self.daemon = pl.use_daemon_threads

		self.set_state(**kwargs)

		if not self.is_alive():
			self.start()

	def critical(self, *args, **kwargs):
		self.pl.critical(prefix=self.__class__.__name__, *args, **kwargs)

	def exception(self, *args, **kwargs):
		self.pl.exception(prefix=self.__class__.__name__, *args, **kwargs)

	def info(self, *args, **kwargs):
		self.pl.info(prefix=self.__class__.__name__, *args, **kwargs)

	def error(self, *args, **kwargs):
		self.pl.error(prefix=self.__class__.__name__, *args, **kwargs)

	def warn(self, *args, **kwargs):
		self.pl.warn(prefix=self.__class__.__name__, *args, **kwargs)

	def debug(self, *args, **kwargs):
		self.pl.debug(prefix=self.__class__.__name__, *args, **kwargs)

	def argspecobjs(self):
		for name in self.argmethods:
			try:
				yield name, getattr(self, name)
			except AttributeError:
				pass

	def additional_args(self):
		return (('interval', self.interval),)

	_omitted_args = {
		'render': (0,),
		'set_state': ('shutdown_event',),
	}

	def omitted_args(self, name, method):
		ret = self._omitted_args.get(name, ())
		if isinstance(getattr(self, name, None), MethodType):
			ret = tuple((i + 1 if isinstance(i, int) else i for i in ret))
		return ret


class KwThreadedSegment(ThreadedSegment):
	update_first = True

	argmethods = ('render', 'set_state', 'key', 'render_one')

	def __init__(self):
		super(KwThreadedSegment, self).__init__()
		self.updated = True
		self.update_value = ({}, set())
		self.write_lock = Lock()
		self.new_queries = []

	@staticmethod
	def key(**kwargs):
		return frozenset(kwargs.items())

	def render(self, update_value, update_first, key=None, after_update=False, **kwargs):
		queries, crashed = update_value
		if key is None:
			key = self.key(**kwargs)
		if key in crashed:
			return self.crashed_value

		try:
			update_state = queries[key][1]
		except KeyError:
			with self.write_lock:
				self.new_queries.append(key)
			if self.do_update_first or self.run_once:
				if after_update:
					self.error('internal error: value was not computed even though update_first was set')
					update_state = None
				else:
					return self.render(
						update_value=self.get_update_value(True),
						update_first=False,
						key=key,
						after_update=True,
						**kwargs
					)
			else:
				update_state = None

		return self.render_one(update_state, **kwargs)

	def update_one(self, crashed, updates, key):
		try:
			updates[key] = (monotonic(), self.compute_state(key))
		except Exception as e:
			self.exception('Exception while computing state for {0!r}: {1}', key, str(e))
			crashed.add(key)
		except KeyboardInterrupt:
			self.warn('Interrupt while computing state for {0!r}', key)
			crashed.add(key)

	def update(self, old_update_value):
		updates = {}
		crashed = set()
		update_value = (updates, crashed)
		queries = old_update_value[0]

		new_queries = self.new_queries
		with self.write_lock:
			self.new_queries = []

		for key, (last_query_time, state) in queries.items():
			if last_query_time < monotonic() < last_query_time + self.interval:
				updates[key] = (last_query_time, state)
			else:
				self.update_one(crashed, updates, key)

		for key in new_queries:
			self.update_one(crashed, updates, key)

		return update_value

	def set_state(self, interval=None, update_first=True, shutdown_event=None, **kwargs):
		self.set_interval(interval)
		self.do_update_first = update_first and self.update_first
		self.shutdown_event = shutdown_event or Event()

	@staticmethod
	def render_one(update_state, **kwargs):
		return update_state

	_omitted_args = {
		'render': ('update_value', 'key', 'after_update'),
		'set_state': ('shutdown_event',),
		'render_one': (0,),
	}



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import sys

from time import sleep
from subprocess import check_call
from glob import glob1
from traceback import print_exc

from argparse import ArgumentParser

from powerline.lib.dict import updated

from tests.modules.lib.terminal import (ExpectProcess, MutableDimensions,
                                        do_terminal_tests, get_env)
from tests.modules import PowerlineTestSuite


TEST_ROOT = os.path.abspath(os.environ['TEST_ROOT'])


def get_parser():
	parser = ArgumentParser()
	parser.add_argument('--type', action='store')
	parser.add_argument('--client', action='store')
	parser.add_argument('--binding', action='store')
	parser.add_argument('args', action='append')
	return parser


BINDING_OPTIONS = {
	'dash': {
		'cmd': 'dash',
		'args': ['-i'],
		'init': [
			'. "$ROOT/tests/test_in_vterm/shell/inits/dash"',
		],
	},
}


def main(argv):
	script_args = get_parser().parse_args(argv)

	vterm_path = os.path.join(TEST_ROOT, 'path')

	env = get_env(vterm_path, TEST_ROOT)
	env['ROOT'] = os.path.abspath('.')
	env['TEST_ROOT'] = TEST_ROOT
	env['TEST_TYPE'] = script_args.type
	env['TEST_CLIENT'] = script_args.client
	env['LANG'] = 'en_US.UTF_8'
	env['_POWERLINE_RUNNING_SHELL_TESTS'] = (
		'ee5bcdc6-b749-11e7-9456-50465d597777')

	dim = MutableDimensions(rows=50, cols=200)

	binding_opts = BINDING_OPTIONS[script_args.binding]

	cmd = os.path.join(vterm_path, binding_opts['cmd'])
	args = binding_opts['args']

	def gen_init(binding):
		def init(p):
			for line in binding_opts['init']:
				p.send(line + '\n')
			sleep(1)

		return init

	def gen_feed(line):
		def feed(p):
			p.send(line + '\n')
			sleep(0.1)

		return feed

	base_attrs = {
		((255, 204,0), (204, 51, 0), 0, 0, 0): 'H',
		((204, 51, 0), (0, 102, 153), 0, 0, 0): 'sHU',
		((255, 255, 255), (0, 102, 153), 1, 0, 0): 'U',
		((0, 102, 153), (44, 44, 44), 0, 0, 0): 'sUB',
		((199, 199, 199), (44, 44, 44), 0, 0, 0): 'B',
		((44, 44, 44), (88, 88, 88), 0, 0, 0): 'sBD',
		((199, 199, 199), (88, 88, 88), 0, 0, 0): 'D',
		((144, 144, 144), (88, 88, 88), 0, 0, 0): 'sD',
		((221, 221, 221), (88, 88, 88), 1, 0, 0): 'C',
		((88, 88, 88), (0, 0, 0), 0, 0, 0): 'sDN',
		((240, 240, 240), (0, 0, 0), 0, 0, 0): 'N',
		((0, 102, 153), (51, 153, 204), 0, 0, 0): 'sUE',
		((255, 255, 255), (51, 153, 204), 0, 0, 0): 'E',
		((51, 153, 204), (44, 44, 44), 0, 0, 0): 'sEB',
	}

	tests = (
		{
			'expected_result': (
				'{H:  hostname }{sHU: }'
				'{U:user }{sUB: }'
				'{B: BRANCH }{sBD: }'
				'{D:… }{sD: }{D:tmp }{sD: }{D:vshells }{sD: }{C:3rd }{sDN: }'
				'{N:}',
				base_attrs,
			),
			'prep_cb': gen_init(script_args.binding),
		},
		{
			'expected_result': (
				'{H:  hostname }{sHU: }'
				'{U:user }{sUB: }'
				'{B: BRANCH }{sBD: }'
				'{D:… }{sD: }{D:vshells }{sD: }{D:3rd }{sD: }{C:.git }{sDN: }'
				'{N:}',
				base_attrs
			),
			'prep_cb': gen_feed('cd .git'),
		},
		{
			'expected_result': (
				'{H:  hostname }{sHU: }'
				'{U:user }{sUB: }'
				'{B: BRANCH }{sBD: }'
				'{D:… }{sD: }{D:tmp }{sD: }{D:vshells }{sD: }{C:3rd }{sDN: }'
				'{N:}',
				base_attrs,
			),
			'prep_cb': gen_feed('cd ..'),
		},
		{
			'expected_result': (
				'{H:  hostname }{sHU: }'
				'{U:user }{sUE: }'
				'{E:(e) some-venv }{sEB: }'
				'{B: BRANCH }{sBD: }'
				'{D:… }{sD: }{D:tmp }{sD: }{D:vshells }{sD: }{C:3rd }{sDN: }'
				'{N:}',
				base_attrs,
			),
			'prep_cb': gen_feed('set_virtual_env some-venv'),
		},
	)

	with PowerlineTestSuite('shell') as suite:
		return do_terminal_tests(
			tests=tests,
			cmd=cmd,
			dim=dim,
			args=args,
			env=env,
			cwd=TEST_ROOT,
			suite=suite,
		)


if __name__ == '__main__':
	if main(sys.argv[1:]):
		raise SystemExit(0)
	else:
		raise SystemExit(1)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import re
import os
import socket

from powerline.lib.url import urllib_read
from powerline.lib.threaded import ThreadedSegment, KwThreadedSegment
from powerline.lib.monotonic import monotonic
from powerline.lib.humanize_bytes import humanize_bytes
from powerline.segments import with_docstring
from powerline.theme import requires_segment_info


@requires_segment_info
def hostname(pl, segment_info, only_if_ssh=False, exclude_domain=False):
	'''Return the current hostname.

	:param bool only_if_ssh:
		only return the hostname if currently in an SSH session
	:param bool exclude_domain:
		return the hostname without domain if there is one
	'''
	if (
		segment_info['environ'].get('_POWERLINE_RUNNING_SHELL_TESTS')
		== 'ee5bcdc6-b749-11e7-9456-50465d597777'
	):
		return 'hostname'
	if only_if_ssh and not segment_info['environ'].get('SSH_CLIENT'):
		return None
	if exclude_domain:
		return socket.gethostname().split('.')[0]
	return socket.gethostname()


def _external_ip(query_url='http://ipv4.icanhazip.com/'):
	return urllib_read(query_url).strip()


class ExternalIpSegment(ThreadedSegment):
	interval = 300

	def set_state(self, query_url='http://ipv4.icanhazip.com/', **kwargs):
		self.query_url = query_url
		super(ExternalIpSegment, self).set_state(**kwargs)

	def update(self, old_ip):
		return _external_ip(query_url=self.query_url)

	def render(self, ip, **kwargs):
		if not ip:
			return None
		return [{'contents': ip, 'divider_highlight_group': 'background:divider'}]


external_ip = with_docstring(ExternalIpSegment(),
'''Return external IP address.

:param str query_url:
	URI to query for IP address, should return only the IP address as a text string

	Suggested URIs:

	* http://ipv4.icanhazip.com/
	* http://ipv6.icanhazip.com/
	* http://icanhazip.com/ (returns IPv6 address if available, else IPv4)

Divider highlight group used: ``background:divider``.
''')


try:
	import netifaces
except ImportError:
	def internal_ip(pl, interface='auto', ipv=4):
		return None
else:
	_interface_starts = {
		'eth':      10,  # Regular ethernet adapters         : eth1
		'enp':      10,  # Regular ethernet adapters, Gentoo : enp2s0
		'en':       10,  # OS X                              : en0 
		'ath':       9,  # Atheros WiFi adapters             : ath0
		'wlan':      9,  # Other WiFi adapters               : wlan1
		'wlp':       9,  # Other WiFi adapters, Gentoo       : wlp5s0
		'teredo':    1,  # miredo interface                  : teredo
		'lo':      -10,  # Loopback interface                : lo
		'docker':   -5,  # Docker bridge interface           : docker0
		'vmnet':    -5,  # VMWare bridge interface           : vmnet1
		'vboxnet':  -5,  # VirtualBox bridge interface       : vboxnet0
	}

	_interface_start_re = re.compile(r'^([a-z]+?)(\d|$)')

	def _interface_key(interface):
		match = _interface_start_re.match(interface)
		if match:
			try:
				base = _interface_starts[match.group(1)] * 100
			except KeyError:
				base = 500
			if match.group(2):
				return base - int(match.group(2))
			else:
				return base
		else:
			return 0

	def internal_ip(pl, interface='auto', ipv=4):
		family = netifaces.AF_INET6 if ipv == 6 else netifaces.AF_INET
		if interface == 'auto':
			try:
				interface = next(iter(sorted(netifaces.interfaces(), key=_interface_key, reverse=True)))
			except StopIteration:
				pl.info('No network interfaces found')
				return None
		elif interface == 'default_gateway':
			try:
				interface = netifaces.gateways()['default'][family][1]
			except KeyError:
				pl.info('No default gateway found for IPv{0}', ipv)
				return None
		addrs = netifaces.ifaddresses(interface)
		try:
			return addrs[family][0]['addr']
		except (KeyError, IndexError):
			pl.info("No IPv{0} address found for interface {1}", ipv, interface)
			return None


internal_ip = with_docstring(internal_ip,
'''Return internal IP address

Requires ``netifaces`` module to work properly.

:param str interface:
	Interface on which IP will be checked. Use ``auto`` to automatically 
	detect interface. In this case interfaces with lower numbers will be 
	preferred over interfaces with similar names. Order of preference based on 
	names:

	#. ``eth`` and ``enp`` followed by number or the end of string.
	#. ``ath``, ``wlan`` and ``wlp`` followed by number or the end of string.
	#. ``teredo`` followed by number or the end of string.
	#. Any other interface that is not ``lo*``.
	#. ``lo`` followed by number or the end of string.

	Use ``default_gateway`` to detect the interface based on the machine's
	`default gateway <https://en.wikipedia.org/wiki/Default_gateway>`_ (i.e.,
	the router to which it is connected).

:param int ipv:
	4 or 6 for ipv4 and ipv6 respectively, depending on which IP address you 
	need exactly.
''')


try:
	import psutil

	def _get_bytes(interface):
		try:
			io_counters = psutil.net_io_counters(pernic=True)
		except AttributeError:
			io_counters = psutil.network_io_counters(pernic=True)
		if_io = io_counters.get(interface)
		if not if_io:
			return None
		return if_io.bytes_recv, if_io.bytes_sent

	def _get_interfaces():
		try:
			io_counters = psutil.net_io_counters(pernic=True)
		except AttributeError:
			io_counters = psutil.network_io_counters(pernic=True)
		for interface, data in io_counters.items():
			if data:
				yield interface, data.bytes_recv, data.bytes_sent
except ImportError:
	def _get_bytes(interface):
		with open('/sys/class/net/{interface}/statistics/rx_bytes'.format(interface=interface), 'rb') as file_obj:
			rx = int(file_obj.read())
		with open('/sys/class/net/{interface}/statistics/tx_bytes'.format(interface=interface), 'rb') as file_obj:
			tx = int(file_obj.read())
		return (rx, tx)

	def _get_interfaces():
		for interface in os.listdir('/sys/class/net'):
			x = _get_bytes(interface)
			if x is not None:
				yield interface, x[0], x[1]


class NetworkLoadSegment(KwThreadedSegment):
	interfaces = {}
	replace_num_pat = re.compile(r'[a-zA-Z]+')

	@staticmethod
	def key(interface='auto', **kwargs):
		return interface

	def compute_state(self, interface):
		if interface == 'auto':
			proc_exists = getattr(self, 'proc_exists', None)
			if proc_exists is None:
				proc_exists = self.proc_exists = os.path.exists('/proc/net/route')
			if proc_exists:
				# Look for default interface in routing table
				with open('/proc/net/route', 'rb') as f:
					for line in f.readlines():
						parts = line.split()
						if len(parts) > 1:
							iface, destination = parts[:2]
							if not destination.replace(b'0', b''):
								interface = iface.decode('utf-8')
								break
			if interface == 'auto':
				# Choose interface with most total activity, excluding some
				# well known interface names
				interface, total = 'eth0', -1
				for name, rx, tx in _get_interfaces():
					base = self.replace_num_pat.match(name)
					if None in (base, rx, tx) or base.group() in ('lo', 'vmnet', 'sit'):
						continue
					activity = rx + tx
					if activity > total:
						total = activity
						interface = name

		try:
			idata = self.interfaces[interface]
			try:
				idata['prev'] = idata['last']
			except KeyError:
				pass
		except KeyError:
			idata = {}
			if self.run_once:
				idata['prev'] = (monotonic(), _get_bytes(interface))
				self.shutdown_event.wait(self.interval)
			self.interfaces[interface] = idata

		idata['last'] = (monotonic(), _get_bytes(interface))
		return idata.copy()

	def render_one(self, idata, recv_format='DL {value:>8}', sent_format='UL {value:>8}', suffix='B/s', si_prefix=False, **kwargs):
		if not idata or 'prev' not in idata:
			return None

		t1, b1 = idata['prev']
		t2, b2 = idata['last']
		measure_interval = t2 - t1

		if None in (b1, b2):
			return None

		r = []
		for i, key in zip((0, 1), ('recv', 'sent')):
			format = locals()[key + '_format']
			try:
				value = (b2[i] - b1[i]) / measure_interval
			except ZeroDivisionError:
				self.warn('Measure interval zero.')
				value = 0
			max_key = key + '_max'
			is_gradient = max_key in kwargs
			hl_groups = ['network_load_' + key, 'network_load']
			if is_gradient:
				hl_groups[:0] = (group + '_gradient' for group in hl_groups)
			r.append({
				'contents': format.format(value=humanize_bytes(value, suffix, si_prefix)),
				'divider_highlight_group': 'network_load:divider',
				'highlight_groups': hl_groups,
			})
			if is_gradient:
				max = kwargs[max_key]
				if value >= max:
					r[-1]['gradient_level'] = 100
				else:
					r[-1]['gradient_level'] = value * 100.0 / max

		return r


network_load = with_docstring(NetworkLoadSegment(),
'''Return the network load.

Uses the ``psutil`` module if available for multi-platform compatibility,
falls back to reading
:file:`/sys/class/net/{interface}/statistics/{rx,tx}_bytes`.

:param str interface:
	Network interface to measure (use the special value "auto" to have powerline 
	try to auto-detect the network interface).
:param str suffix:
	String appended to each load string.
:param bool si_prefix:
	Use SI prefix, e.g. MB instead of MiB.
:param str recv_format:
	Format string that determines how download speed should look like. Receives 
	``value`` as argument.
:param str sent_format:
	Format string that determines how upload speed should look like. Receives 
	``value`` as argument.
:param float recv_max:
	Maximum number of received bytes per second. Is only used to compute
	gradient level.
:param float sent_max:
	Maximum number of sent bytes per second. Is only used to compute gradient
	level.

Divider highlight group used: ``network_load:divider``.

Highlight groups used: ``network_load_sent_gradient`` (gradient) or ``network_load_recv_gradient`` (gradient) or ``network_load_gradient`` (gradient), ``network_load_sent`` or ``network_load_recv`` or ``network_load``.
''')



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import shutil
import os

from time import sleep
from functools import partial
from errno import ENOENT

from powerline.lib.watcher import create_file_watcher, create_tree_watcher, INotifyError
from powerline.lib.watcher.uv import UvNotFound
from powerline import get_fallback_logger
from powerline.lib.monotonic import monotonic

from tests.modules import TestCase, SkipTest


INOTIFY_DIR = 'inotify' + os.path.basename(os.environ.get('PYTHON', ''))


def clear_dir(dir):
	for root, dirs, files in list(os.walk(dir, topdown=False)):
		for f in files:
			os.remove(os.path.join(root, f))
		for d in dirs:
			os.rmdir(os.path.join(root, d))


def set_watcher_tests(l):
	byte_tests = (('bytes', True), ('unicode', False))

	for btn, use_bytes in byte_tests:
		def test_inotify_file_watcher(self, use_bytes=use_bytes):
			try:
				w = create_file_watcher(pl=get_fallback_logger(), watcher_type='inotify')
			except INotifyError:
				raise SkipTest('This test is not suitable for a stat based file watcher')
			self.do_test_file_watcher(w, use_bytes)

		def test_uv_file_watcher(self, use_bytes=use_bytes):
			raise SkipTest('Uv watcher tests are not stable')
			try:
				w = create_file_watcher(pl=get_fallback_logger(), watcher_type='uv')
			except UvNotFound:
				raise SkipTest('Pyuv is not available')
			self.do_test_file_watcher(w, use_bytes)

		def test_inotify_tree_watcher(self, use_bytes=use_bytes):
			try:
				tw = create_tree_watcher(get_fallback_logger(), watcher_type='inotify')
			except INotifyError:
				raise SkipTest('INotify is not available')
			self.do_test_tree_watcher(tw, use_bytes)

		def test_uv_tree_watcher(self, use_bytes=use_bytes):
			raise SkipTest('Uv watcher tests are not stable')
			try:
				tw = create_tree_watcher(get_fallback_logger(), 'uv')
			except UvNotFound:
				raise SkipTest('Pyuv is not available')
			self.do_test_tree_watcher(tw, use_bytes)

		def test_inotify_file_watcher_is_watching(self, use_bytes=use_bytes):
			try:
				w = create_file_watcher(pl=get_fallback_logger(), watcher_type='inotify')
			except INotifyError:
				raise SkipTest('INotify is not available')
			self.do_test_file_watcher_is_watching(w, use_bytes)

		def test_stat_file_watcher_is_watching(self, use_bytes=use_bytes):
			w = create_file_watcher(pl=get_fallback_logger(), watcher_type='stat')
			self.do_test_file_watcher_is_watching(w, use_bytes)

		def test_uv_file_watcher_is_watching(self, use_bytes=use_bytes):
			try:
				w = create_file_watcher(pl=get_fallback_logger(), watcher_type='uv')
			except UvNotFound:
				raise SkipTest('Pyuv is not available')
			self.do_test_file_watcher_is_watching(w, use_bytes)

		for wt in ('uv', 'inotify'):
			l['test_{0}_file_watcher_{1}'.format(wt, btn)] = locals()['test_{0}_file_watcher'.format(wt)]
			l['test_{0}_tree_watcher_{1}'.format(wt, btn)] = locals()['test_{0}_tree_watcher'.format(wt)]
			l['test_{0}_file_watcher_is_watching_{1}'.format(wt, btn)] = (
				locals()['test_{0}_file_watcher_is_watching'.format(wt)])
		l['test_{0}_file_watcher_is_watching_{1}'.format('stat', btn)] = (
			locals()['test_{0}_file_watcher_is_watching'.format('stat')])


class TestFilesystemWatchers(TestCase):
	def do_test_for_change(self, watcher, path):
		st = monotonic()
		while monotonic() - st < 1:
			if watcher(path):
				return
			sleep(0.1)
		self.fail('The change to {0} was not detected'.format(path))

	def do_test_file_watcher(self, w, use_bytes=False):
		try:
			f1, f2, f3 = map(lambda x: os.path.join(INOTIFY_DIR, 'file%d' % x), (1, 2, 3))
			ne = os.path.join(INOTIFY_DIR, 'notexists')
			if use_bytes:
				f1 = f1.encode('utf-8')
				f2 = f2.encode('utf-8')
				f3 = f3.encode('utf-8')
				ne = ne.encode('utf-8')
			with open(f1, 'wb'):
				with open(f2, 'wb'):
					with open(f3, 'wb'):
						pass
			self.assertRaises(OSError, w, ne)
			self.assertTrue(w(f1))
			self.assertTrue(w(f2))
			os.utime(f1, None), os.utime(f2, None)
			self.do_test_for_change(w, f1)
			self.do_test_for_change(w, f2)
			# Repeat once
			os.utime(f1, None), os.utime(f2, None)
			self.do_test_for_change(w, f1)
			self.do_test_for_change(w, f2)
			# Check that no false changes are reported
			self.assertFalse(w(f1), 'Spurious change detected')
			self.assertFalse(w(f2), 'Spurious change detected')
			# Check that open the file with 'w' triggers a change
			with open(f1, 'wb'):
				with open(f2, 'wb'):
					pass
			self.do_test_for_change(w, f1)
			self.do_test_for_change(w, f2)
			# Check that writing to a file with 'a' triggers a change
			with open(f1, 'ab') as f:
				f.write(b'1')
			self.do_test_for_change(w, f1)
			# Check that deleting a file registers as a change
			os.unlink(f1)
			self.do_test_for_change(w, f1)
			# Test that changing the inode of a file does not cause it to stop
			# being watched
			os.rename(f3, f2)
			self.do_test_for_change(w, f2)
			self.assertFalse(w(f2), 'Spurious change detected')
			os.utime(f2, None)
			self.do_test_for_change(w, f2)
		finally:
			clear_dir(INOTIFY_DIR)

	def do_test_tree_watcher(self, tw, use_bytes=False):
		try:
			inotify_dir = INOTIFY_DIR
			subdir = os.path.join(inotify_dir, 'subdir')
			t1 = os.path.join(inotify_dir, 'tree1')
			ts1 = os.path.join(subdir, 'tree1')
			suffix = '1'
			f = os.path.join(subdir, 'f')
			if use_bytes:
				inotify_dir = inotify_dir.encode('utf-8')
				subdir = subdir.encode('utf-8')
				t1 = t1.encode('utf-8')
				ts1 = ts1.encode('utf-8')
				suffix = suffix.encode('utf-8')
				f = f.encode('utf-8')
			os.mkdir(subdir)
			try:
				if tw.watch(inotify_dir).is_dummy:
					raise SkipTest('No tree watcher available')
			except UvNotFound:
				raise SkipTest('Pyuv is not available')
			except INotifyError:
				raise SkipTest('INotify is not available')
			self.assertTrue(tw(inotify_dir))
			self.assertFalse(tw(inotify_dir))
			changed = partial(self.do_test_for_change, tw, inotify_dir)
			open(t1, 'w').close()
			changed()
			open(ts1, 'w').close()
			changed()
			os.unlink(ts1)
			changed()
			os.rmdir(subdir)
			changed()
			os.mkdir(subdir)
			changed()
			os.rename(subdir, subdir + suffix)
			changed()
			shutil.rmtree(subdir + suffix)
			changed()
			os.mkdir(subdir)
			open(f, 'w').close()
			changed()
			with open(f, 'a') as s:
				s.write(' ')
			changed()
			os.rename(f, f + suffix)
			changed()
		finally:
			clear_dir(inotify_dir)

	def do_test_file_watcher_is_watching(self, w, use_bytes=False):
		try:
			f1, f2, f3 = map(lambda x: os.path.join(INOTIFY_DIR, 'file%d' % x), (1, 2, 3))
			ne = os.path.join(INOTIFY_DIR, 'notexists')
			if use_bytes:
				f1 = f1.encode('utf-8')
				f2 = f2.encode('utf-8')
				f3 = f3.encode('utf-8')
				ne = ne.encode('utf-8')
			with open(f1, 'wb'):
				with open(f2, 'wb'):
					with open(f3, 'wb'):
						pass
			self.assertRaises(OSError, w, ne)
			try:
				w(ne)
			except OSError as e:
				self.assertEqual(e.errno, ENOENT)
			self.assertTrue(w(f1))
			self.assertFalse(w.is_watching(ne))
			self.assertTrue(w.is_watching(f1))
			self.assertFalse(w.is_watching(f2))
		finally:
			clear_dir(INOTIFY_DIR)

	set_watcher_tests(locals())


old_cwd = None


def setUpModule():
	global old_cwd
	old_cwd = os.getcwd()
	os.chdir(os.path.dirname(os.path.dirname(__file__)))
	os.mkdir(INOTIFY_DIR)


def tearDownModule():
	shutil.rmtree(INOTIFY_DIR)
	os.chdir(old_cwd)


if __name__ == '__main__':
	from tests.modules import main
	main()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import ctypes

from collections import namedtuple

from powerline.lib.unicode import unicode, unichr, tointiter


Dimensions = namedtuple('Dimensions', ('rows', 'cols'))


class CTypesFunction(object):
	def __init__(self, library, name, rettype, args):
		self.name = name
		self.prototype = ctypes.CFUNCTYPE(rettype, *[
			arg[1] for arg in args
		])
		self.args = args
		self.func = self.prototype((name, library), tuple((
			(1, arg[0]) for arg in args
		)))

	def __call__(self, *args, **kwargs):
		return self.func(*args, **kwargs)

	def __repr__(self):
		return '{cls}(<library>, {name!r}, {rettype!r}, {args!r})'.format(
			cls=self.__class__.__name__,
			**self.__dict__
		)


class CTypesLibraryFuncsCollection(object):
	def __init__(self, lib, **kwargs):
		self.lib = lib
		library_loader = ctypes.LibraryLoader(ctypes.CDLL)
		library = library_loader.LoadLibrary(lib)
		self.library = library
		for name, args in kwargs.items():
			self.__dict__[name] = CTypesFunction(library, name, *args)


class VTermPos_s(ctypes.Structure):
	_fields_ = (
		('row', ctypes.c_int),
		('col', ctypes.c_int),
	)


class VTermColor_s(ctypes.Structure):
	_fields_ = (
		('red', ctypes.c_uint8),
		('green', ctypes.c_uint8),
		('blue', ctypes.c_uint8),
	)


class VTermScreenCellAttrs_s(ctypes.Structure):
	_fields_ = (
		('bold', ctypes.c_uint, 1),
		('underline', ctypes.c_uint, 2),
		('italic', ctypes.c_uint, 1),
		('blink', ctypes.c_uint, 1),
		('reverse', ctypes.c_uint, 1),
		('strike', ctypes.c_uint, 1),
		('font', ctypes.c_uint, 4),
		('dwl', ctypes.c_uint, 1),
		('dhl', ctypes.c_uint, 2),
	)


VTERM_MAX_CHARS_PER_CELL = 6


class VTermScreenCell_s(ctypes.Structure):
	_fields_ = (
		('chars', ctypes.ARRAY(ctypes.c_uint32, VTERM_MAX_CHARS_PER_CELL)),
		('width', ctypes.c_char),
		('attrs', VTermScreenCellAttrs_s),
		('fg', VTermColor_s),
		('bg', VTermColor_s),
	)


VTerm_p = ctypes.c_void_p
VTermScreen_p = ctypes.c_void_p


def get_functions(lib):
	return CTypesLibraryFuncsCollection(
		lib,
		vterm_new=(VTerm_p, (
			('rows', ctypes.c_int),
			('cols', ctypes.c_int)
		)),
		vterm_obtain_screen=(VTermScreen_p, (('vt', VTerm_p),)),
		vterm_set_size=(None, (
			('vt', VTerm_p),
			('rows', ctypes.c_int),
			('cols', ctypes.c_int)
		)),
		vterm_screen_reset=(None, (
			('screen', VTermScreen_p),
			('hard', ctypes.c_int)
		)),
		vterm_input_write=(ctypes.c_size_t, (
			('vt', VTerm_p),
			('bytes', ctypes.POINTER(ctypes.c_char)),
			('size', ctypes.c_size_t),
		)),
		vterm_screen_get_cell=(ctypes.c_int, (
			('screen', VTermScreen_p),
			('pos', VTermPos_s),
			('cell', ctypes.POINTER(VTermScreenCell_s))
		)),
		vterm_free=(None, (('vt', VTerm_p),)),
		vterm_set_utf8=(None, (('vt', VTerm_p), ('is_utf8', ctypes.c_int))),
	)


class VTermColor(object):
	__slots__ = ('red', 'green', 'blue')

	def __init__(self, color):
		self.red = color.red
		self.green = color.green
		self.blue = color.blue

	@property
	def color_key(self):
		return (self.red, self.green, self.blue)


class VTermScreenCell(object):
	def __init__(self, vtsc):
		for field in VTermScreenCellAttrs_s._fields_:
			field_name = field[0]
			setattr(self, field_name, getattr(vtsc.attrs, field_name))
		self.text = ''.join((
			unichr(vtsc.chars[i]) for i in range(VTERM_MAX_CHARS_PER_CELL)
		)).rstrip('\x00')
		self.width = next(tointiter(vtsc.width))
		self.fg = VTermColor(vtsc.fg)
		self.bg = VTermColor(vtsc.bg)
		self.cell_properties_key = (
			self.fg.color_key,
			self.bg.color_key,
			self.bold,
			self.underline,
			self.italic,
		)


class VTermScreen(object):
	def __init__(self, functions, screen):
		self.functions = functions
		self.screen = screen

	def __getitem__(self, position):
		pos = VTermPos_s(*position)
		cell = VTermScreenCell_s()
		ret = self.functions.vterm_screen_get_cell(self.screen, pos, cell)
		if ret != 1:
			raise ValueError('vterm_screen_get_cell returned {0}'.format(ret))
		return VTermScreenCell(cell)

	def reset(self, hard):
		self.functions.vterm_screen_reset(self.screen, int(bool(hard)))


class VTerm(object):
	def __init__(self, lib, dim):
		self.functions = get_functions(lib)
		self.vt = self.functions.vterm_new(dim.rows, dim.cols)
		self.functions.vterm_set_utf8(self.vt, 1)
		self.vtscreen = VTermScreen(self.functions, self.functions.vterm_obtain_screen(self.vt))
		self.vtscreen.reset(True)

	def push(self, data):
		if isinstance(data, unicode):
			data = data.encode('utf-8')
		return self.functions.vterm_input_write(self.vt, data, len(data))

	def resize(self, dim):
		self.functions.vterm_set_size(self.vt, dim.rows, dim.cols)

	def __del__(self):
		try:
			self.functions.vterm_free(self.vt)
		except AttributeError:
			pass



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import re
import sys
import subprocess
import shlex

from powerline.config import POWERLINE_ROOT, TMUX_CONFIG_DIRECTORY
from powerline.lib.config import ConfigLoader
from powerline import generate_config_finder, load_config, create_logger, finish_common_config
from powerline.shell import ShellPowerline
from powerline.lib.shell import which
from powerline.bindings.tmux import (TmuxVersionInfo, run_tmux_command, set_tmux_environment, get_tmux_version,
                                     source_tmux_file)
from powerline.lib.encoding import get_preferred_output_encoding
from powerline.renderers.tmux import attrs_to_tmux_attrs
from powerline.commands.main import finish_args


CONFIG_FILE_NAME = re.compile(r'powerline_tmux_(?P<major>\d+)\.(?P<minor>\d+)(?P<suffix>[a-z]+)?(?:_(?P<mod>plus|minus))?\.conf')
CONFIG_MATCHERS = {
	None: (lambda a, b: a.major == b.major and a.minor == b.minor),
	'plus': (lambda a, b: a[:2] <= b[:2]),
	'minus': (lambda a, b: a[:2] >= b[:2]),
}
CONFIG_PRIORITY = {
	None: 3,
	'plus': 2,
	'minus': 1,
}


def list_all_tmux_configs():
	'''List all version-specific tmux configuration files'''
	for root, dirs, files in os.walk(TMUX_CONFIG_DIRECTORY):
		dirs[:] = ()
		for fname in files:
			match = CONFIG_FILE_NAME.match(fname)
			if match:
				assert match.group('suffix') is None
				yield (
					os.path.join(root, fname),
					CONFIG_MATCHERS[match.group('mod')],
					CONFIG_PRIORITY[match.group('mod')],
					TmuxVersionInfo(
						int(match.group('major')),
						int(match.group('minor')),
						match.group('suffix'),
					),
				)


def get_tmux_configs(version):
	'''Get tmux configuration suffix given parsed tmux version

	:param TmuxVersionInfo version: Parsed tmux version.
	'''
	for fname, matcher, priority, file_version in list_all_tmux_configs():
		if matcher(file_version, version):
			yield (fname, priority + file_version.minor * 10 + file_version.major * 10000)


def source_tmux_files(pl, args, tmux_version=None, source_tmux_file=source_tmux_file):
	'''Source relevant version-specific tmux configuration files

	Files are sourced in the following order:
	* First relevant files with older versions are sourced.
	* If files for same versions are to be sourced then first _minus files are 
	  sourced, then _plus files and then files without _minus or _plus suffixes.
	'''
	tmux_version = tmux_version or get_tmux_version(pl)
	source_tmux_file(os.path.join(TMUX_CONFIG_DIRECTORY, 'powerline-base.conf'))
	for fname, priority in sorted(get_tmux_configs(tmux_version), key=(lambda v: v[1])):
		source_tmux_file(fname)
	if not os.environ.get('POWERLINE_COMMAND'):
		cmd = deduce_command()
		if cmd:
			set_tmux_environment('POWERLINE_COMMAND', deduce_command(), remove=False)
	try:
		run_tmux_command('refresh-client')
	except subprocess.CalledProcessError:
		# On tmux-2.0 this command may fail for whatever reason. Since it is 
		# critical just ignore the failure.
		pass


class EmptyArgs(object):
	def __init__(self, ext, config_path):
		self.ext = [ext]
		self.side = 'left'
		self.config_path = None

	def __getattr__(self, attr):
		return None


def init_tmux_environment(pl, args, set_tmux_environment=set_tmux_environment):
	'''Initialize tmux environment from tmux configuration
	'''
	powerline = ShellPowerline(finish_args(None, os.environ, EmptyArgs('tmux', args.config_path)))
	# TODO Move configuration files loading out of Powerline object and use it 
	# directly
	powerline.update_renderer()
	# FIXME Use something more stable then `theme_kwargs`
	colorscheme = powerline.renderer_options['theme_kwargs']['colorscheme']

	def get_highlighting(group):
		return colorscheme.get_highlighting([group], None)

	for varname, highlight_group in (
		('_POWERLINE_BACKGROUND_COLOR', 'background'),
		('_POWERLINE_ACTIVE_WINDOW_STATUS_COLOR', 'active_window_status'),
		('_POWERLINE_WINDOW_STATUS_COLOR', 'window_status'),
		('_POWERLINE_ACTIVITY_STATUS_COLOR', 'activity_status'),
		('_POWERLINE_BELL_STATUS_COLOR', 'bell_status'),
		('_POWERLINE_WINDOW_COLOR', 'window'),
		('_POWERLINE_WINDOW_DIVIDER_COLOR', 'window:divider'),
		('_POWERLINE_WINDOW_CURRENT_COLOR', 'window:current'),
		('_POWERLINE_WINDOW_NAME_COLOR', 'window_name'),
		('_POWERLINE_SESSION_COLOR', 'session'),
	):
		highlight = get_highlighting(highlight_group)
		set_tmux_environment(varname, powerline.renderer.hlstyle(**highlight)[2:-1])
	for varname, prev_group, next_group in (
		('_POWERLINE_WINDOW_CURRENT_HARD_DIVIDER_COLOR', 'window', 'window:current'),
		('_POWERLINE_WINDOW_CURRENT_HARD_DIVIDER_NEXT_COLOR', 'window:current', 'window'),
		('_POWERLINE_SESSION_HARD_DIVIDER_NEXT_COLOR', 'session', 'background'),
	):
		prev_highlight = get_highlighting(prev_group)
		next_highlight = get_highlighting(next_group)
		set_tmux_environment(
			varname,
			powerline.renderer.hlstyle(
				fg=prev_highlight['bg'],
				bg=next_highlight['bg'],
				attrs=0,
			)[2:-1]
		)
	for varname, attr, group in (
		('_POWERLINE_ACTIVE_WINDOW_FG', 'fg', 'active_window_status'),
		('_POWERLINE_WINDOW_STATUS_FG', 'fg', 'window_status'),
		('_POWERLINE_ACTIVITY_STATUS_FG', 'fg', 'activity_status'),
		('_POWERLINE_ACTIVITY_STATUS_ATTR', 'attrs', 'activity_status'),
		('_POWERLINE_BELL_STATUS_FG', 'fg', 'bell_status'),
		('_POWERLINE_BELL_STATUS_ATTR', 'attrs', 'bell_status'),
		('_POWERLINE_BACKGROUND_FG', 'fg', 'background'),
		('_POWERLINE_BACKGROUND_BG', 'bg', 'background'),
		('_POWERLINE_SESSION_FG', 'fg', 'session'),
		('_POWERLINE_SESSION_BG', 'bg', 'session'),
		('_POWERLINE_SESSION_ATTR', 'attrs', 'session'),
		('_POWERLINE_SESSION_PREFIX_FG', 'fg', 'session:prefix'),
		('_POWERLINE_SESSION_PREFIX_BG', 'bg', 'session:prefix'),
		('_POWERLINE_SESSION_PREFIX_ATTR', 'attrs', 'session:prefix'),
	):
		if attr == 'attrs':
			attrs = attrs_to_tmux_attrs(get_highlighting(group)[attr])
			set_tmux_environment(varname, ']#['.join(attrs))
			set_tmux_environment(varname + '_LEGACY', (','.join(
				# Tmux-1.6 does not accept no… attributes in 
				# window-status-…-attr options.
				(attr for attr in attrs if not attr.startswith('no')))
				# But it does not support empty attributes as well.
				or 'none'))
		else:
			if powerline.common_config['term_truecolor']:
				set_tmux_environment(varname, '#{0:06x}'.format(get_highlighting(group)[attr][1]))
			else:
				set_tmux_environment(varname, 'colour' + str(get_highlighting(group)[attr][0]))

	left_dividers = powerline.renderer.theme.dividers['left']
	set_tmux_environment('_POWERLINE_LEFT_HARD_DIVIDER', left_dividers['hard'])
	set_tmux_environment('_POWERLINE_LEFT_SOFT_DIVIDER', left_dividers['soft'])
	set_tmux_environment('_POWERLINE_LEFT_HARD_DIVIDER_SPACES', (
		' ' * powerline.renderer.strwidth(left_dividers['hard'])))


TMUX_VAR_RE = re.compile('\$(_POWERLINE_\w+)')


def tmux_setup(pl, args):
	tmux_environ = {}
	tmux_version = get_tmux_version(pl)

	def set_tmux_environment_nosource(varname, value, remove=True):
		tmux_environ[varname] = value

	def replace_cb(match):
		return tmux_environ[match.group(1)]

	def replace_env(s):
		return TMUX_VAR_RE.subn(replace_cb, s)[0]

	def source_tmux_file_nosource(fname):
		with open(fname) as fd:
			for line in fd:
				if line.startswith('#') or line == '\n':
					continue
				args = shlex.split(line)
				args = [args[0]] + [replace_env(arg) for arg in args[1:]]
				run_tmux_command(*args)

	if args.source is None:
		args.source = tmux_version < (1, 9)

	if args.source:
		ste = set_tmux_environment
		stf = source_tmux_file
	else:
		ste = set_tmux_environment_nosource
		stf = source_tmux_file_nosource

	init_tmux_environment(pl, args, set_tmux_environment=ste)
	source_tmux_files(pl, args, tmux_version=tmux_version, source_tmux_file=stf)


def get_main_config(args):
	find_config_files = generate_config_finder()
	config_loader = ConfigLoader(run_once=True)
	return load_config('config', find_config_files, config_loader)


def create_powerline_logger(args):
	config = get_main_config(args)
	common_config = finish_common_config(get_preferred_output_encoding(), config['common'])
	logger, pl, get_module_attr = create_logger(common_config)
	return pl


def check_command(cmd):
	if which(cmd):
		return cmd


def deduce_command():
	'''Deduce which command to use for ``powerline``

	Candidates:

	* ``powerline``. Present only when installed system-wide.
	* ``{powerline_root}/scripts/powerline``. Present after ``pip install -e`` 
	  was run and C client was compiled (in this case ``pip`` does not install 
	  binary file).
	* ``{powerline_root}/client/powerline.sh``. Useful when ``sh``, ``sed`` and 
	  ``socat`` are present, but ``pip`` or ``setup.py`` was not run.
	* ``{powerline_root}/client/powerline.py``. Like above, but when one of 
	  ``sh``, ``sed`` and ``socat`` was not present.
	* ``powerline-render``. Should not really ever be used.
	* ``{powerline_root}/scripts/powerline-render``. Same.
	'''
	return (
		None
		or check_command('powerline')
		or check_command(os.path.join(POWERLINE_ROOT, 'scripts', 'powerline'))
		or ((which('sh') and which('sed') and which('socat'))
			and check_command(os.path.join(POWERLINE_ROOT, 'client', 'powerline.sh')))
		or check_command(os.path.join(POWERLINE_ROOT, 'client', 'powerline.py'))
		or check_command('powerline-render')
		or check_command(os.path.join(POWERLINE_ROOT, 'scripts', 'powerline-render'))
	)


def shell_command(pl, args):
	cmd = deduce_command()
	if cmd:
		print(cmd)
	else:
		sys.exit(1)


def uses(pl, args):
	component = args.component
	if not component:
		raise ValueError('Must specify component')
	shell = args.shell
	template = 'POWERLINE_NO_{shell}_{component}'
	for sh in (shell, 'shell') if shell else ('shell'):
		varname = template.format(shell=sh.upper(), component=component.upper())
		if os.environ.get(varname):
			sys.exit(1)
	config = get_main_config(args)
	if component in config.get('ext', {}).get('shell', {}).get('components', ('tmux', 'prompt')):
		sys.exit(0)
	else:
		sys.exit(1)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from powerline.lib.unicode import unicode


def havemarks(*args, **kwargs):
	origin = kwargs.get('origin', '')
	for i, v in enumerate(args):
		if not hasattr(v, 'mark'):
			raise AssertionError('Value #{0}/{1} ({2!r}) has no attribute `mark`'.format(origin, i, v))
		if isinstance(v, dict):
			for key, val in v.items():
				havemarks(key, val, origin=(origin + '[' + unicode(i) + ']/' + unicode(key)))
		elif isinstance(v, list):
			havemarks(*v, origin=(origin + '[' + unicode(i) + ']'))



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import imp
import sys


class Pl(object):
	def __init__(self):
		self.exceptions = []
		self.errors = []
		self.warns = []
		self.debugs = []
		self.infos = []
		self.prefix = None
		self.use_daemon_threads = True

	for meth in ('error', 'warn', 'debug', 'exception', 'info'):
		exec((
			'def {0}(self, msg, *args, **kwargs):\n'
			'	self.{0}s.append((kwargs.get("prefix") or self.prefix, msg, args, kwargs))\n'
		).format(meth))

	def __nonzero__(self):
		return bool(self.exceptions or self.errors or self.warns)

	__bool__ = __nonzero__


class Args(object):
	theme_override = {}
	config_override = {}
	config_path = None
	ext = ['shell']
	renderer_module = None

	def __init__(self, **kwargs):
		self.__dict__.update(kwargs)


def urllib_read(query_url):
	if query_url.startswith('http://ipv'):
		if query_url.startswith('http://ipv4.icanhazip.com'):
			return '127.0.0.1'
		elif query_url.startswith('http://ipv6.icanhazip.com'):
			return '2001:4801:7818:6:abc5:ba2c:ff10:275f'

	elif query_url.startswith('https://freegeoip.app/json/'):
		return '{"ip":"82.145.55.16","country_code":"DE","country_name":"Germany","region_code":"NI","region_name":"Lower Saxony","city":"Meppen","zip_code":"49716","time_zone":"Europe/Berlin","latitude":52.6833,"longitude":7.3167,"metro_code":0}'
	elif query_url.startswith('http://geoip.nekudo.com/api/'):
		return '{"city":"Meppen","country":{"name":"Germany", "code":"DE"},"location":{"accuracy_radius":100,"latitude":52.6833,"longitude":7.3167,"time_zone":"Europe\/Berlin"},"ip":"82.145.55.16"}'
	elif query_url.startswith('http://query.yahooapis.com/v1/public/'):
		if 'Meppen' in query_url or '52.6833' in query_url:
			return r'{"query":{"count":1,"created":"2016-05-13T19:43:18Z","lang":"en-US","results":{"channel":{"units":{"distance":"mi","pressure":"in","speed":"mph","temperature":"C"},"title":"Yahoo! Weather - Meppen, NI, DE","link":"http://us.rd.yahoo.com/dailynews/rss/weather/Country__Country/*https://weather.yahoo.com/country/state/city-674836/","description":"Yahoo! Weather for Meppen, NI, DE","language":"en-us","lastBuildDate":"Fri, 13 May 2016 09:43 PM CEST","ttl":"60","location":{"city":"Meppen","country":"Germany","region":" NI"},"wind":{"chill":"55","direction":"350","speed":"25"},"atmosphere":{"humidity":"57","pressure":"1004.0","rising":"0","visibility":"16.1"},"astronomy":{"sunrise":"5:35 am","sunset":"9:21 pm"},"image":{"title":"Yahoo! Weather","width":"142","height":"18","link":"http://weather.yahoo.com","url":"http://l.yimg.com/a/i/brand/purplelogo//uh/us/news-wea.gif"},"item":{"title":"Conditions for Meppen, NI, DE at 08:00 PM CEST","lat":"52.68993","long":"7.29115","link":"http://us.rd.yahoo.com/dailynews/rss/weather/Country__Country/*https://weather.yahoo.com/country/state/city-674836/","pubDate":"Fri, 13 May 2016 08:00 PM CEST","condition":{"code":"23","date":"Fri, 13 May 2016 08:00 PM CEST","temp":"14","text":"Breezy"},"forecast":[{"code":"30","date":"13 May 2016","day":"Fri","high":"71","low":"48","text":"Partly Cloudy"},{"code":"28","date":"14 May 2016","day":"Sat","high":"54","low":"44","text":"Mostly Cloudy"},{"code":"11","date":"15 May 2016","day":"Sun","high":"55","low":"43","text":"Showers"},{"code":"28","date":"16 May 2016","day":"Mon","high":"54","low":"42","text":"Mostly Cloudy"},{"code":"28","date":"17 May 2016","day":"Tue","high":"57","low":"43","text":"Mostly Cloudy"},{"code":"12","date":"18 May 2016","day":"Wed","high":"62","low":"45","text":"Rain"},{"code":"28","date":"19 May 2016","day":"Thu","high":"63","low":"48","text":"Mostly Cloudy"},{"code":"28","date":"20 May 2016","day":"Fri","high":"67","low":"50","text":"Mostly Cloudy"},{"code":"30","date":"21 May 2016","day":"Sat","high":"71","low":"50","text":"Partly Cloudy"},{"code":"30","date":"22 May 2016","day":"Sun","high":"74","low":"54","text":"Partly Cloudy"}],"description":"<![CDATA[<img src=\"http://l.yimg.com/a/i/us/we/52/23.gif\"/>\n<BR />\n<b>Current Conditions:</b>\n<BR />Breezy\n<BR />\n<BR />\n<b>Forecast:</b>\n<BR /> Fri - Partly Cloudy. High: 71Low: 48\n<BR /> Sat - Mostly Cloudy. High: 54Low: 44\n<BR /> Sun - Showers. High: 55Low: 43\n<BR /> Mon - Mostly Cloudy. High: 54Low: 42\n<BR /> Tue - Mostly Cloudy. High: 57Low: 43\n<BR />\n<BR />\n<a href=\"http://us.rd.yahoo.com/dailynews/rss/weather/Country__Country/*https://weather.yahoo.com/country/state/city-674836/\">Full Forecast at Yahoo! Weather</a>\n<BR />\n<BR />\n(provided by <a href=\"http://www.weather.com\" >The Weather Channel</a>)\n<BR />\n]]>","guid":{"isPermaLink":"false"}}}}}}'
		elif 'Moscow' in query_url:
			return r'{"query":{"count":1,"created":"2016-05-13T19:47:01Z","lang":"en-US","results":{"channel":{"units":{"distance":"mi","pressure":"in","speed":"mph","temperature":"C"},"title":"Yahoo! Weather - Moscow, Moscow Federal City, RU","link":"http://us.rd.yahoo.com/dailynews/rss/weather/Country__Country/*https://weather.yahoo.com/country/state/city-2122265/","description":"Yahoo! Weather for Moscow, Moscow Federal City, RU","language":"en-us","lastBuildDate":"Fri, 13 May 2016 10:47 PM MSK","ttl":"60","location":{"city":"Moscow","country":"Russia","region":" Moscow Federal City"},"wind":{"chill":"45","direction":"80","speed":"11"},"atmosphere":{"humidity":"52","pressure":"993.0","rising":"0","visibility":"16.1"},"astronomy":{"sunrise":"4:19 am","sunset":"8:34 pm"},"image":{"title":"Yahoo! Weather","width":"142","height":"18","link":"http://weather.yahoo.com","url":"http://l.yimg.com/a/i/brand/purplelogo//uh/us/news-wea.gif"},"item":{"title":"Conditions for Moscow, Moscow Federal City, RU at 09:00 PM MSK","lat":"55.741638","long":"37.605061","link":"http://us.rd.yahoo.com/dailynews/rss/weather/Country__Country/*https://weather.yahoo.com/country/state/city-2122265/","pubDate":"Fri, 13 May 2016 09:00 PM MSK","condition":{"code":"33","date":"Fri, 13 May 2016 09:00 PM MSK","temp":"9","text":"Mostly Clear"},"forecast":[{"code":"30","date":"13 May 2016","day":"Fri","high":"62","low":"41","text":"Partly Cloudy"},{"code":"30","date":"14 May 2016","day":"Sat","high":"64","low":"43","text":"Partly Cloudy"},{"code":"30","date":"15 May 2016","day":"Sun","high":"63","low":"44","text":"Partly Cloudy"},{"code":"12","date":"16 May 2016","day":"Mon","high":"60","low":"47","text":"Rain"},{"code":"12","date":"17 May 2016","day":"Tue","high":"64","low":"48","text":"Rain"},{"code":"28","date":"18 May 2016","day":"Wed","high":"67","low":"48","text":"Mostly Cloudy"},{"code":"12","date":"19 May 2016","day":"Thu","high":"68","low":"49","text":"Rain"},{"code":"39","date":"20 May 2016","day":"Fri","high":"66","low":"50","text":"Scattered Showers"},{"code":"39","date":"21 May 2016","day":"Sat","high":"69","low":"49","text":"Scattered Showers"},{"code":"30","date":"22 May 2016","day":"Sun","high":"73","low":"50","text":"Partly Cloudy"}],"description":"<![CDATA[<img src=\"http://l.yimg.com/a/i/us/we/52/33.gif\"/>\n<BR />\n<b>Current Conditions:</b>\n<BR />Mostly Clear\n<BR />\n<BR />\n<b>Forecast:</b>\n<BR /> Fri - Partly Cloudy. High: 62Low: 41\n<BR /> Sat - Partly Cloudy. High: 64Low: 43\n<BR /> Sun - Partly Cloudy. High: 63Low: 44\n<BR /> Mon - Rain. High: 60Low: 47\n<BR /> Tue - Rain. High: 64Low: 48\n<BR />\n<BR />\n<a href=\"http://us.rd.yahoo.com/dailynews/rss/weather/Country__Country/*https://weather.yahoo.com/country/state/city-2122265/\">Full Forecast at Yahoo! Weather</a>\n<BR />\n<BR />\n(provided by <a href=\"http://www.weather.com\" >The Weather Channel</a>)\n<BR />\n]]>","guid":{"isPermaLink":"false"}}}}}}'
	elif query_url.startswith('https://api.openweathermap.org/data/2.5/'):
		if 'Meppen' in query_url or '52.6833' in query_url:
			return r'{"coord":{"lon":7.29,"lat":52.69},"weather":[{"id":800,"main":"Clear","description":"clear sky","icon":"01n"}],"base":"stations","main":{"temp":293.15,"feels_like":295.16,"temp_min":293.15,"temp_max":295.37,"pressure":1018,"humidity":77},"visibility":10000,"wind":{"speed":1.12,"deg":126},"clouds":{"all":0},"dt":1600196220,"sys":{"type":1,"id":1871,"country":"DE","sunrise":1600146332,"sunset":1600191996},"timezone":7200,"id":2871845,"name":"Meppen","cod":200}'
		elif 'Moscow' in query_url:
			return r'{"coord":{"lon":37.62,"lat":55.75},"weather":[{"id":800,"main":"Clear","description":"clear sky","icon":"01n"}],"base":"stations","main":{"temp":283.15,"feels_like":280.78,"temp_min":283.15,"temp_max":284.26,"pressure":1019,"humidity":71},"visibility":10000,"wind":{"speed":3,"deg":330},"clouds":{"all":0},"dt":1600196224,"sys":{"type":1,"id":9029,"country":"RU","sunrise":1600138909,"sunset":1600184863},"timezone":10800,"id":524901,"name":"Moscow","cod":200}'

	raise NotImplementedError


class Process(object):
	def __init__(self, output, err):
		self.output = output
		self.err = err

	def communicate(self):
		return self.output, self.err


class ModuleReplace(object):
	def __init__(self, name, new):
		self.name = name
		self.new = new

	def __enter__(self):
		self.old = sys.modules.get(self.name)
		if not self.old:
			try:
				self.old = __import__(self.name)
			except ImportError:
				pass
		sys.modules[self.name] = self.new

	def __exit__(self, *args):
		if self.old:
			sys.modules[self.name] = self.old
		else:
			sys.modules.pop(self.name)


def replace_module(name, new=None, **kwargs):
	if not new:
		new = new_module(name, **kwargs)
	return ModuleReplace(name, new)


def new_module(name, **kwargs):
	module = imp.new_module(name)
	for k, v in kwargs.items():
		setattr(module, k, v)
	return module


class AttrReplace(object):
	def __init__(self, obj, *args):
		self.obj = obj
		self.attrs = args[::2]
		self.new = args[1::2]

	def __enter__(self):
		self.old = {}
		for i, attr in enumerate(self.attrs):
			try:
				self.old[i] = getattr(self.obj, attr)
			except AttributeError:
				pass
		for attr, new in zip(self.attrs, self.new):
			setattr(self.obj, attr, new)

	def __exit__(self, *args):
		for i, attr in enumerate(self.attrs):
			try:
				old = self.old[i]
			except KeyError:
				delattr(self.obj, attr)
			else:
				setattr(self.obj, attr, old)


replace_attr = AttrReplace


def replace_module_module(module, name, **kwargs):
	return replace_attr(module, name, new_module(name, **kwargs))


class ItemReplace(object):
	def __init__(self, d, key, new, r=None):
		self.key = key
		self.new = new
		self.d = d
		self.r = r

	def __enter__(self):
		self.old = self.d.get(self.key)
		self.d[self.key] = self.new
		return self.r

	def __exit__(self, *args):
		if self.old is None:
			try:
				self.d.pop(self.key)
			except KeyError:
				pass
		else:
			self.d[self.key] = self.old


def replace_item(d, key, new):
	return ItemReplace(d, key, new, d)


def replace_env(key, new, environ=None, **kwargs):
	r = kwargs.copy()
	r['environ'] = environ or {}
	return ItemReplace(r['environ'], key, new, r)


class PowerlineSingleTest(object):
	def __init__(self, suite, name):
		self.suite = suite
		self.name = name

	def __enter__(self):
		return self

	def __exit__(self, exc_type, exc_value, traceback):
		if exc_type is not None:
			self.exception('Exception while running test: {0!r}'.format(
				exc_value))

	def fail(self, message, allow_failure=False):
		return self.suite.fail(self.name, message, allow_failure)

	def exception(self, message, allow_failure=False):
		return self.suite.exception(self.name, message, allow_failure)



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import socket
import sys
import codecs
import platform
import re


test_root = os.environ['TEST_ROOT']
test_type = sys.argv[1]
test_client = sys.argv[2]
shell = sys.argv[3]
fname = os.path.join(test_root, '.'.join((shell, test_type, test_client, 'full.log')))
new_fname = os.path.join(test_root, '.'.join((shell, test_type, test_client, 'log')))
pid_fname = os.path.join(test_root, '3rd', 'pid')

is_pypy = platform.python_implementation() == 'PyPy'


try:
	with open(pid_fname, 'r') as P:
		pid = P.read().strip()
except IOError:
	pid = None
hostname = socket.gethostname()
user = os.environ['USER']

REFS_RE = re.compile(r'^\[\d+ refs\]\n')
IPYPY_DEANSI_RE = re.compile(r'\033(?:\[(?:\?\d+[lh]|[^a-zA-Z]+[a-ln-zA-Z])|[=>])')
ZSH_HL_RE = re.compile(r'\033\[\?\d+[hl]')

start_str = 'cd "$TEST_ROOT"/3rd'
if shell == 'pdb':
	start_str = 'class Foo(object):'

with codecs.open(fname, 'r', encoding='utf-8') as R:
	with codecs.open(new_fname, 'w', encoding='utf-8') as W:
		found_cd = False
		i = -1
		for line in (R if shell != 'fish' else R.read().split('\n')):
			i += 1
			if not found_cd:
				found_cd = (start_str in line)
				continue
			if 'true is the last line' in line:
				break
			line = line.translate({
				ord('\r'): None
			})
			if REFS_RE.match(line):
				continue
			line = line.replace(hostname, 'HOSTNAME')
			line = line.replace(user, 'USER')
			if pid is not None:
				line = line.replace(pid, 'PID')
			if shell == 'zsh':
				line = line.replace('\033[0m\033[23m\033[24m\033[J', '')
				line = ZSH_HL_RE.subn('', line)[0]
			elif shell == 'fish':
				res = ''
				try:
					while line.index('\033[0;'):
						start = line.index('\033[0;')
						end = line.index('\033[0m', start)
						res += line[start:end + 4] + '\n'
						line = line[end + 4:]
				except ValueError:
					pass
				line = res
			elif shell == 'tcsh':
				try:
					start = line.index('\033[0;')
					end = line.index(' ', start)
					line = line[start:end] + '\n'
				except ValueError:
					line = ''
			elif shell == 'mksh':
				# Output is different in travis: on my machine I see full 
				# command, in travis it is truncated just after `true`.
				if line.startswith('[1] + Terminated'):
					line = '[1] + Terminated bash -c ...\n'
			elif shell == 'dash':
				# Position of this line is not stable: it may go both before and 
				# after the next line
				if line.startswith('[1] + Terminated'):
					continue
			elif shell == 'ipython' and is_pypy:
				try:
					end_idx = line.rindex('\033[0m')
					try:
						idx = line[:end_idx].rindex('\033[1;1H')
					except ValueError:
						idx = line[:end_idx].rindex('\033[?25h')
					line = line[idx + len('\033[1;1H'):]
				except ValueError:
					pass
				try:
					data_end_idx = line.rindex('\033[1;1H')
					line = line[:data_end_idx] + '\n'
				except ValueError:
					pass
				if line == '\033[1;1H\n':
					continue
				was_empty = line == '\n'
				line = IPYPY_DEANSI_RE.subn('', line)[0]
				if line == '\n' and not was_empty:
					line = ''
			elif shell == 'rc':
				if line == 'read() failed: Connection reset by peer\n':
					line = ''
			elif shell == 'pdb':
				if is_pypy:
					if line == '\033[?1h\033=\033[?25l\033[1A\n':
						line = ''
					line = IPYPY_DEANSI_RE.subn('', line)[0]
					if line == '\n':
						line = ''
				if line.startswith(('>',)):
					line = ''
				elif line == '-> self.quitting = 1\n':
					line = '-> self.quitting = True\n'
				elif line == '\n':
					line = ''
				if line == '-> self.quitting = True\n':
					break
			W.write(line)



# vim:fileencoding=utf-8:noet
def nop(_):
	pass


class Foo(object):
	def __init__(self):
		nop('__init__')
		self.bar()
		self.baz()
		self.bra()

	@classmethod
	def bar(cls):
		nop(cls.__name__)

	@staticmethod
	def baz():
		nop(1)

	def bra(self):
		nop(self.__class__.__name__)


def brah():
	nop('brah')


f = Foo()
Foo.bar()
Foo.baz()
Foo.bra(f)

f.bar()
f.baz()
f.bra()

brah()



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from time import sleep
from copy import deepcopy
from functools import wraps

from tests.modules import TestCase
from tests.modules.lib.config_mock import get_powerline, add_watcher_events, UT


config = {
	'config': {
		'common': {
			'interval': 0,
			'watcher': 'test',
		},
		'ext': {
			'test': {
				'theme': 'default',
				'colorscheme': 'default',
			},
		},
	},
	'colors': {
		'colors': {
			"col1": 1,
			"col2": 2,
			"col3": 3,
			"col4": 4,
		},
		'gradients': {
		},
	},
	'colorschemes/test/default': {
		'groups': {
			'str1': {'fg': 'col1', 'bg': 'col2', 'attrs': ['bold']},
			'str2': {'fg': 'col3', 'bg': 'col4', 'attrs': ['underline']},
		},
	},
	'colorschemes/test/2': {
		'groups': {
			'str1': {'fg': 'col2', 'bg': 'col3', 'attrs': ['bold']},
			'str2': {'fg': 'col1', 'bg': 'col4', 'attrs': ['underline']},
		},
	},
	'themes/test/default': {
		'segments': {
			"left": [
				{
					"type": "string",
					"contents": "s",
					"highlight_groups": ["str1"],
				},
				{
					"type": "string",
					"contents": "g",
					"highlight_groups": ["str2"],
				},
			],
			"right": [
			],
		},
	},
	'themes/' + UT: {
		'dividers': {
			"left": {
				"hard": ">>",
				"soft": ">",
			},
			"right": {
				"hard": "<<",
				"soft": "<",
			},
		},
		'spaces': 0,
	},
	'themes/other': {
		'dividers': {
			"left": {
				"hard": ">>",
				"soft": ">",
			},
			"right": {
				"hard": "<<",
				"soft": "<",
			},
		},
		'spaces': 1,
	},
	'themes/test/2': {
		'segments': {
			"left": [
				{
					"type": "string",
					"contents": "t",
					"highlight_groups": ["str1"],
				},
				{
					"type": "string",
					"contents": "b",
					"highlight_groups": ["str2"],
				},
			],
			"right": [
			],
		},
	},
}


def with_new_config(func):
	@wraps(func)
	def f(self):
		return func(self, deepcopy(config))
	return f


class TestConfigReload(TestCase):
	def assertAccessEvents(self, p, *args):
		events = set()
		for event in args:
			if ':' not in event:
				events.add('check:' + event)
				events.add('load:' + event)
			else:
				events.add(event)
		self.assertEqual(set(p._pop_events()), events)

	@with_new_config
	def test_noreload(self, config):
		with get_powerline(config, run_once=True) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')
			config['config']['common']['spaces'] = 1
			add_watcher_events(p, 'config', wait=False, interval=0.05)
			# When running once thread should not start
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p)
			self.assertEqual(p.logger._pop_msgs(), [])

	@with_new_config
	def test_reload_main(self, config):
		with get_powerline(config, run_once=False) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['config']['common']['default_top_theme'] = 'other'
			add_watcher_events(p, 'config')
			p.render()
			self.assertEqual(p.render(), '<1 2 1> s <2 4 False>>><3 4 4>g <4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'themes/other', 'check:themes/test/__main__', 'themes/test/default')
			self.assertEqual(p.logger._pop_msgs(), [])

			config['config']['ext']['test']['theme'] = 'nonexistent'
			add_watcher_events(p, 'config')
			self.assertEqual(p.render(), '<1 2 1> s <2 4 False>>><3 4 4>g <4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'check:themes/test/nonexistent', 'themes/other', 'check:themes/test/__main__')
			# It should normally handle file missing error
			self.assertEqual(p.logger._pop_msgs(), [
				'exception:test:powerline:Failed to load theme: themes/test/__main__',
				'exception:test:powerline:Failed to load theme: themes/test/nonexistent',
				'exception:test:powerline:Failed to create renderer: themes/test/nonexistent'
			])

			config['config']['ext']['test']['theme'] = 'default'
			add_watcher_events(p, 'config')
			self.assertEqual(p.render(), '<1 2 1> s <2 4 False>>><3 4 4>g <4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'themes/test/default', 'themes/other', 'check:themes/test/__main__')
			self.assertEqual(p.logger._pop_msgs(), [])

			config['config']['ext']['test']['colorscheme'] = 'nonexistent'
			add_watcher_events(p, 'config')
			self.assertEqual(p.render(), '<1 2 1> s <2 4 False>>><3 4 4>g <4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'check:colorschemes/nonexistent', 'check:colorschemes/test/__main__', 'check:colorschemes/test/nonexistent')
			# It should normally handle file missing error
			self.assertEqual(p.logger._pop_msgs(), [
				'exception:test:powerline:Failed to load colorscheme: colorschemes/nonexistent',
				'exception:test:powerline:Failed to load colorscheme: colorschemes/test/__main__',
				'exception:test:powerline:Failed to load colorscheme: colorschemes/test/nonexistent',
				'exception:test:powerline:Failed to create renderer: colorschemes/test/nonexistent'
			])

			config['config']['ext']['test']['colorscheme'] = '2'
			add_watcher_events(p, 'config')
			self.assertEqual(p.render(), '<2 3 1> s <3 4 False>>><1 4 4>g <4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'check:colorschemes/2', 'check:colorschemes/test/__main__', 'colorschemes/test/2')
			self.assertEqual(p.logger._pop_msgs(), [])

			config['config']['ext']['test']['theme'] = '2'
			add_watcher_events(p, 'config')
			self.assertEqual(p.render(), '<2 3 1> t <3 4 False>>><1 4 4>b <4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'themes/test/2', 'themes/other', 'check:themes/test/__main__')
			self.assertEqual(p.logger._pop_msgs(), [])

			self.assertEqual(p.renderer.local_themes, None)
			config['config']['ext']['test']['local_themes'] = 'something'
			add_watcher_events(p, 'config')
			self.assertEqual(p.render(), '<2 3 1> t <3 4 False>>><1 4 4>b <4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config')
			self.assertEqual(p.logger._pop_msgs(), [])
			self.assertEqual(p.renderer.local_themes, 'something')

	@with_new_config
	def test_reload_unexistent(self, config):
		with get_powerline(config, run_once=False) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['config']['ext']['test']['colorscheme'] = 'nonexistentraise'
			add_watcher_events(p, 'config')
			# It may appear that p.logger._pop_msgs() is called after given 
			# exception is added to the mesagges, but before config_loader 
			# exception was added (this one: 
			# “exception:test:config_loader:Error while running condition 
			# function for key colorschemes/test/nonexistentraise: 
			# fcf:colorschemes/test/nonexistentraise”).
			# sleep(0.1)
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			# For colorschemes/{test/,}*raise find_config_file raises 
			# IOError, but it does not do so for check:colorschemes/test/__main__, 
			# so powerline is trying to load this, but not other 
			# colorschemes/*
			self.assertAccessEvents(p, 'config', 'check:colorschemes/test/__main__', 'check:colorschemes/nonexistentraise', 'check:colorschemes/test/nonexistentraise')
			self.assertIn('exception:test:powerline:Failed to create renderer: fcf:colorschemes/test/nonexistentraise', p.logger._pop_msgs())

			config['colorschemes/nonexistentraise'] = {}
			config['colorschemes/test/nonexistentraise'] = {
				'groups': {
					'str1': {'fg': 'col1', 'bg': 'col3', 'attrs': ['bold']},
					'str2': {'fg': 'col2', 'bg': 'col4', 'attrs': ['underline']},
				},
			}
			while not p._will_create_renderer():
				sleep(0.1)
			self.assertEqual(p.render(), '<1 3 1> s<3 4 False>>><2 4 4>g<4 False False>>><None None None>')
			# Same as above
			self.assertAccessEvents(p, 'colorschemes/nonexistentraise', 'colorschemes/test/nonexistentraise', 'check:colorschemes/test/__main__')
			self.assertEqual(p.logger._pop_msgs(), [])

	@with_new_config
	def test_reload_colors(self, config):
		with get_powerline(config, run_once=False) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['colors']['colors']['col1'] = 5
			add_watcher_events(p, 'colors')
			self.assertEqual(p.render(), '<5 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'colors')
			self.assertEqual(p.logger._pop_msgs(), [])

	@with_new_config
	def test_reload_colorscheme(self, config):
		with get_powerline(config, run_once=False) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['colorschemes/test/default']['groups']['str1']['bg'] = 'col3'
			add_watcher_events(p, 'colorschemes/test/default')
			self.assertEqual(p.render(), '<1 3 1> s<3 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default')
			self.assertEqual(p.logger._pop_msgs(), [])

	@with_new_config
	def test_reload_theme(self, config):
		with get_powerline(config, run_once=False) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['themes/test/default']['segments']['left'][0]['contents'] = 'col3'
			add_watcher_events(p, 'themes/test/default')
			self.assertEqual(p.render(), '<1 2 1> col3<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')
			self.assertEqual(p.logger._pop_msgs(), [])

	@with_new_config
	def test_reload_top_theme(self, config):
		with get_powerline(config, run_once=False) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['themes/' + UT]['dividers']['left']['hard'] = '|>'
			add_watcher_events(p, 'themes/' + UT)
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>|><3 4 4>g<4 False False>|><None None None>')
			self.assertAccessEvents(p, 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')
			self.assertEqual(p.logger._pop_msgs(), [])

	@with_new_config
	def test_reload_theme_main(self, config):
		config['config']['common']['interval'] = None
		with get_powerline(config, run_once=False) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['themes/test/default']['segments']['left'][0]['contents'] = 'col3'
			add_watcher_events(p, 'themes/test/default', wait=False)
			self.assertEqual(p.render(), '<1 2 1> col3<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')
			self.assertEqual(p.logger._pop_msgs(), [])
			self.assertTrue(p._watcher._calls)

	@with_new_config
	def test_run_once_no_theme_reload(self, config):
		config['config']['common']['interval'] = None
		with get_powerline(config, run_once=True) as p:
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p, 'config', 'colors', 'check:colorschemes/default', 'check:colorschemes/test/__main__', 'colorschemes/test/default', 'themes/test/default', 'themes/' + UT, 'check:themes/test/__main__')

			config['themes/test/default']['segments']['left'][0]['contents'] = 'col3'
			add_watcher_events(p, 'themes/test/default', wait=False)
			self.assertEqual(p.render(), '<1 2 1> s<2 4 False>>><3 4 4>g<4 False False>>><None None None>')
			self.assertAccessEvents(p)
			self.assertEqual(p.logger._pop_msgs(), [])


if __name__ == '__main__':
	from tests.modules import main
	main()



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import sys

from time import sleep
from subprocess import check_call
from glob import glob1
from traceback import print_exc

from powerline.lib.dict import updated

from tests.modules.lib.terminal import (ExpectProcess, MutableDimensions,
                                        do_terminal_tests, get_env)
from tests.modules import PowerlineTestSuite


TEST_ROOT = os.path.abspath(os.environ['TEST_ROOT'])


def main(attempts=3):
	vterm_path = os.path.join(TEST_ROOT, 'path')

	vim_exe = os.path.join(vterm_path, 'vim')

	env = get_env(vterm_path, TEST_ROOT)
	env['ROOT'] = os.path.abspath('.')

	dim = MutableDimensions(rows=50, cols=200)

	vimrc = os.path.join(TEST_ROOT, 'init.vim')
	vimrc_contents = '''
		set laststatus=2
		set runtimepath=$ROOT/powerline/bindings/vim
	'''
	with open(vimrc, 'w') as vd:
		vd.write(vimrc_contents)

	base_attrs = {
		(( 64,  64, 255), (0, 0, 0), 0, 0, 0): 'NT',  # NonText
		((240, 240, 240), (0, 0, 0), 0, 0, 0): 'N',   # Normal
	}

	args = [
		'-u', vimrc,
		'-i', 'NONE',
	]

	def feed(p):
		p.send(':echo strtrans(eval(&statusline[2:]))\n')

	tests = (
	)

	with PowerlineTestSuite('vim') as suite:
		return do_terminal_tests(
			tests=tests,
			cmd=vim_exe,
			dim=dim,
			args=args,
			env=env,
			cwd=TEST_ROOT,
			suite=suite,
		)


if __name__ == '__main__':
	if main():
		raise SystemExit(0)
	else:
		raise SystemExit(1)



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import argparse
import os
import re

from time import sleep
from subprocess import check_call
from io import BytesIO

import pexpect


def get_argparser(ArgumentParser=argparse.ArgumentParser):
	parser = ArgumentParser(description='Run powerline shell test using pexpect')
	parser.add_argument('--wait-for-echo', action='store_true', help='Wait until the input is echoed back.')
	parser.add_argument('--type', metavar='TYPE', help='Test type (daemon, nodaemon, …).')
	parser.add_argument('--client', metavar='CLIENT', help='Type of the client used (C, shell, zpython, …).')
	parser.add_argument('--shell', metavar='SHELL', help='Shell name.')
	parser.add_argument('command', nargs=argparse.REMAINDER, metavar='COMMAND',
	                    help='Command to run and its argument.')
	return parser


def main():
	test_root = os.environ['TEST_ROOT']
	parser = get_argparser()
	args = parser.parse_args()

	shell = args.shell or args.command[0]
	test_type = args.type or shell
	test_client = args.client or test_type

	log_file_base = '{0}.{1}.{2}'.format(shell, test_type, test_client)
	full_log_file_name = os.path.join(test_root, '{0}.full.log'.format(log_file_base))

	local_paths = [
		os.path.abspath(os.path.join(test_root, 'path')),
		os.path.abspath('scripts'),
	]

	if test_type == 'fish':
		local_paths += ['/usr/bin', '/bin']

	python_paths = os.environ.get('PYTHONPATH', '')
	if python_paths:
		python_paths = ':' + python_paths
	python_paths = os.path.abspath('.') + python_paths

	environ = {
		'LANG': 'en_US.UTF-8',
		'PATH': os.pathsep.join(local_paths),
		'TERM': 'screen-256color',
		'DIR1': os.environ['DIR1'],
		'DIR2': os.environ['DIR2'],
		'XDG_CONFIG_HOME': os.path.abspath(os.path.join(test_root, 'fish_home')),
		'IPYTHONDIR': os.path.abspath(os.path.join(test_root, 'ipython_home')),
		'PYTHONPATH': python_paths,
		'POWERLINE_CONFIG_OVERRIDES': os.environ.get('POWERLINE_CONFIG_OVERRIDES', ''),
		'POWERLINE_THEME_OVERRIDES': os.environ.get('POWERLINE_THEME_OVERRIDES', ''),
		'POWERLINE_CONFIG_PATHS': os.path.abspath(os.path.join('powerline', 'config_files')),
		'POWERLINE_COMMAND_ARGS': os.environ.get('POWERLINE_COMMAND_ARGS', ''),
		'POWERLINE_COMMAND': os.environ.get('POWERLINE_COMMAND', ''),
		'LD_LIBRARY_PATH': os.environ.get('LD_LIBRARY_PATH', ''),
		'TEST_ROOT': test_root,
	}

	os.environ['PATH'] = environ['PATH']

	if test_type == 'daemon':
		environ['POWERLINE_SHELL_CONTINUATION'] = '1'
		environ['POWERLINE_SHELL_SELECT'] = '1'

	if test_type != 'zpython' and shell == 'zsh':
		environ['POWERLINE_NO_ZSH_ZPYTHON'] = '1'

	sio = BytesIO()

	child = pexpect.spawn(
		args.command[0],
		args.command[1:],
		env=environ,
		logfile=sio,
		timeout=30,
	)
	child.expect(re.compile(b'.*'))
	sleep(0.5)
	child.setwinsize(1, 300)

	with open(os.path.join('tests', 'test_shells', 'inputs', shell), 'rb') as F:
		if not args.wait_for_echo:
			child.send(F.read())
		else:
			for line in F:
				child.send(line)
				sleep(1)
				# TODO Implement something more smart

	with open(full_log_file_name, 'wb') as LF:
		while True:
			try:
				s = child.read_nonblocking(1000)
			except pexpect.TIMEOUT:
				break
			except pexpect.EOF:
				break
			else:
				LF.write(s)

	child.close(force=True)

	check_call([
		os.path.join(test_root, 'path', 'python'),
		os.path.join('tests', 'test_shells', 'postproc.py'),
		test_type, test_client, shell
	])
	pidfile = os.path.join(test_root, '3rd', 'pid')
	if os.path.exists(pidfile):
		os.unlink(pidfile)


if __name__ == '__main__':
	main()



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import argparse

from getpass import getpass

from github import Github


p = argparse.ArgumentParser(description='Powerline release script')
p.add_argument('-u', '--user', type=str, metavar='USER', help='Github username.', required=True)
p.add_argument('-p', '--password', type=str, metavar='PASS', help='Github password. You will be prompted if it is not supplied.')

if __name__ == '__main__':
	args = p.parse_args()
	user = args.user
	password = args.password or getpass('Password for {0}: '.format(user))
	gh = Github(user, password)
	grepo = gh.get_repo('powerline/powerline')
	for pr in grepo.get_pulls():
		if pr.base.ref != 'develop':
			issue = grepo.get_issue(pr.number)
			issue.create_comment('PRs to any branch, but develop, are not accepted.', )
			issue.add_to_labels('s:invalid')
			issue.edit(state='closed')



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import collections
import types

from functools import wraps

from powerline.lint.markedjson.error import MarkedError

from powerline.lint.markedjson import nodes
from powerline.lint.markedjson.markedvalue import gen_marked_value
from powerline.lib.unicode import unicode


def marked(func):
	@wraps(func)
	def f(self, node, *args, **kwargs):
		return gen_marked_value(func(self, node, *args, **kwargs), node.start_mark)
	return f


class ConstructorError(MarkedError):
	pass


class BaseConstructor:
	yaml_constructors = {}

	def __init__(self):
		self.constructed_objects = {}
		self.state_generators = []
		self.deep_construct = False

	def check_data(self):
		# If there are more documents available?
		return self.check_node()

	def get_data(self):
		# Construct and return the next document.
		if self.check_node():
			return self.construct_document(self.get_node())

	def get_single_data(self):
		# Ensure that the stream contains a single document and construct it.
		node = self.get_single_node()
		if node is not None:
			return self.construct_document(node)
		return None

	def construct_document(self, node):
		data = self.construct_object(node)
		while self.state_generators:
			state_generators = self.state_generators
			self.state_generators = []
			for generator in state_generators:
				for dummy in generator:
					pass
		self.constructed_objects = {}
		self.deep_construct = False
		return data

	def construct_object(self, node, deep=False):
		if node in self.constructed_objects:
			return self.constructed_objects[node]
		if deep:
			old_deep = self.deep_construct
			self.deep_construct = True
		constructor = None
		tag_suffix = None
		if node.tag in self.yaml_constructors:
			constructor = self.yaml_constructors[node.tag]
		else:
			raise ConstructorError(None, None, 'no constructor for tag %s' % node.tag)
		if tag_suffix is None:
			data = constructor(self, node)
		else:
			data = constructor(self, tag_suffix, node)
		if isinstance(data, types.GeneratorType):
			generator = data
			data = next(generator)
			if self.deep_construct:
				for dummy in generator:
					pass
			else:
				self.state_generators.append(generator)
		self.constructed_objects[node] = data
		if deep:
			self.deep_construct = old_deep
		return data

	@marked
	def construct_scalar(self, node):
		if not isinstance(node, nodes.ScalarNode):
			raise ConstructorError(
				None, None,
				'expected a scalar node, but found %s' % node.id,
				node.start_mark
			)
		return node.value

	def construct_sequence(self, node, deep=False):
		if not isinstance(node, nodes.SequenceNode):
			raise ConstructorError(
				None, None,
				'expected a sequence node, but found %s' % node.id,
				node.start_mark
			)
		return [
			self.construct_object(child, deep=deep)
			for child in node.value
		]

	@marked
	def construct_mapping(self, node, deep=False):
		if not isinstance(node, nodes.MappingNode):
			raise ConstructorError(
				None, None,
				'expected a mapping node, but found %s' % node.id,
				node.start_mark
			)
		mapping = {}
		for key_node, value_node in node.value:
			key = self.construct_object(key_node, deep=deep)
			if not isinstance(key, collections.abc.Hashable):
				self.echoerr(
					'While constructing a mapping', node.start_mark,
					'found unhashable key', key_node.start_mark
				)
				continue
			elif type(key.value) != unicode:
				self.echoerr(
					'Error while constructing a mapping', node.start_mark,
					'found key that is not a string', key_node.start_mark
				)
				continue
			elif key in mapping:
				self.echoerr(
					'Error while constructing a mapping', node.start_mark,
					'found duplicate key', key_node.start_mark
				)
				continue
			value = self.construct_object(value_node, deep=deep)
			mapping[key] = value
		return mapping

	@classmethod
	def add_constructor(cls, tag, constructor):
		if 'yaml_constructors' not in cls.__dict__:
			cls.yaml_constructors = cls.yaml_constructors.copy()
		cls.yaml_constructors[tag] = constructor


class Constructor(BaseConstructor):
	def construct_scalar(self, node):
		if isinstance(node, nodes.MappingNode):
			for key_node, value_node in node.value:
				if key_node.tag == 'tag:yaml.org,2002:value':
					return self.construct_scalar(value_node)
		return BaseConstructor.construct_scalar(self, node)

	def flatten_mapping(self, node):
		merge = []
		index = 0
		while index < len(node.value):
			key_node, value_node = node.value[index]
			if key_node.tag == 'tag:yaml.org,2002:merge':
				del node.value[index]
				if isinstance(value_node, nodes.MappingNode):
					self.flatten_mapping(value_node)
					merge.extend(value_node.value)
				elif isinstance(value_node, nodes.SequenceNode):
					submerge = []
					for subnode in value_node.value:
						if not isinstance(subnode, nodes.MappingNode):
							raise ConstructorError(
								'while constructing a mapping',
								node.start_mark,
								'expected a mapping for merging, but found %s' % subnode.id,
								subnode.start_mark
							)
						self.flatten_mapping(subnode)
						submerge.append(subnode.value)
					submerge.reverse()
					for value in submerge:
						merge.extend(value)
				else:
					raise ConstructorError(
						'while constructing a mapping',
						node.start_mark,
						('expected a mapping or list of mappings for merging, but found %s' % value_node.id),
						value_node.start_mark
					)
			elif key_node.tag == 'tag:yaml.org,2002:value':
				key_node.tag = 'tag:yaml.org,2002:str'
				index += 1
			else:
				index += 1
		if merge:
			node.value = merge + node.value

	def construct_mapping(self, node, deep=False):
		if isinstance(node, nodes.MappingNode):
			self.flatten_mapping(node)
		return BaseConstructor.construct_mapping(self, node, deep=deep)

	@marked
	def construct_yaml_null(self, node):
		self.construct_scalar(node)
		return None

	@marked
	def construct_yaml_bool(self, node):
		value = self.construct_scalar(node).value
		return bool(value)

	@marked
	def construct_yaml_int(self, node):
		value = self.construct_scalar(node).value
		sign = +1
		if value[0] == '-':
			sign = -1
		if value[0] in '+-':
			value = value[1:]
		if value == '0':
			return 0
		else:
			return sign * int(value)

	@marked
	def construct_yaml_float(self, node):
		value = self.construct_scalar(node).value
		sign = +1
		if value[0] == '-':
			sign = -1
		if value[0] in '+-':
			value = value[1:]
		else:
			return sign * float(value)

	def construct_yaml_str(self, node):
		return self.construct_scalar(node)

	def construct_yaml_seq(self, node):
		data = gen_marked_value([], node.start_mark)
		yield data
		data.extend(self.construct_sequence(node))

	def construct_yaml_map(self, node):
		data = gen_marked_value({}, node.start_mark)
		yield data
		value = self.construct_mapping(node)
		data.update(value)

	def construct_undefined(self, node):
		raise ConstructorError(
			None, None,
			'could not determine a constructor for the tag %r' % node.tag,
			node.start_mark
		)


Constructor.add_constructor(
	'tag:yaml.org,2002:null', Constructor.construct_yaml_null)

Constructor.add_constructor(
	'tag:yaml.org,2002:bool', Constructor.construct_yaml_bool)

Constructor.add_constructor(
	'tag:yaml.org,2002:int', Constructor.construct_yaml_int)

Constructor.add_constructor(
	'tag:yaml.org,2002:float', Constructor.construct_yaml_float)

Constructor.add_constructor(
	'tag:yaml.org,2002:str', Constructor.construct_yaml_str)

Constructor.add_constructor(
	'tag:yaml.org,2002:seq', Constructor.construct_yaml_seq)

Constructor.add_constructor(
	'tag:yaml.org,2002:map', Constructor.construct_yaml_map)

Constructor.add_constructor(
	None, Constructor.construct_undefined)



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys

from powerline.bindings.wm import DEFAULT_UPDATE_INTERVAL
from powerline.bindings.wm.awesome import run


def main():
	try:
		interval = float(sys.argv[1])
	except IndexError:
		interval = DEFAULT_UPDATE_INTERVAL
	run(interval=interval)


if __name__ == '__main__':
	main()



#!/usr/bin/env python
# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import os
import sys
import json

from time import sleep
from subprocess import check_call
from glob import glob1
from traceback import print_exc

from powerline.lib.dict import updated
from powerline.bindings.tmux import get_tmux_version
from powerline import get_fallback_logger

from tests.modules.lib.terminal import (ExpectProcess, MutableDimensions,
                                        do_terminal_tests, get_env)
from tests.modules import PowerlineTestSuite


TEST_ROOT = os.path.abspath(os.environ['TEST_ROOT'])


def tmux_logs_iter(test_dir):
	for tail in glob1(test_dir, '*.log'):
		yield os.path.join(test_dir, tail)


def print_tmux_logs():
	for f in tmux_logs_iter(TEST_ROOT):
		print('_' * 80)
		print(os.path.basename(f) + ':')
		print('=' * 80)
		with open(f, 'r') as fp:
			for line in fp:
				sys.stdout.write(line)
		os.unlink(f)


def get_expected_result(tmux_version,
                        expected_result_old,
                        expected_result_1_7=None,
                        expected_result_1_8=None,
                        expected_result_2_0=None):
	if tmux_version >= (2, 0) and expected_result_2_0:
		return expected_result_2_0
	elif tmux_version >= (1, 8) and expected_result_1_8:
		return expected_result_1_8
	elif tmux_version >= (1, 7) and expected_result_1_7:
		return expected_result_1_7
	else:
		return expected_result_old


def tmux_fin_cb(p, cmd, env):
	try:
		check_call([
			cmd, '-S', env['POWERLINE_TMUX_SOCKET_PATH'], 'kill-server'
		], env=env, cwd=TEST_ROOT)
	except Exception:
		print_exc()
	for f in tmux_logs_iter(TEST_ROOT):
		os.unlink(f)


def main(attempts=3):
	vterm_path = os.path.join(TEST_ROOT, 'path')

	tmux_exe = os.path.join(vterm_path, 'tmux')

	socket_path = os.path.abspath('tmux-socket-{0}'.format(attempts))
	if os.path.exists(socket_path):
		os.unlink(socket_path)

	env = get_env(vterm_path, TEST_ROOT, {
		'POWERLINE_THEME_OVERRIDES': ';'.join((
			key + '=' + json.dumps(val)
			for key, val in (
				('default.segments.right', [{
					'type': 'string',
					'name': 's1',
					'highlight_groups': ['cwd'],
					'priority':50,
				}]),
				('default.segments.left', [{
					'type': 'string',
					'name': 's2',
					'highlight_groups': ['background'],
					'priority':20,
				}]),
				('default.segment_data.s1.contents', 'S1 string here'),
				('default.segment_data.s2.contents', 'S2 string here'),
			)
		)),
		'POWERLINE_TMUX_SOCKET_PATH': socket_path,
	})

	conf_path = os.path.abspath('powerline/bindings/tmux/powerline.conf')
	conf_line = 'source "' + (
		conf_path.replace('\\', '\\\\').replace('"', '\\"')) + '"\n'
	conf_file = os.path.realpath(os.path.join(TEST_ROOT, 'tmux.conf'))
	with open(conf_file, 'w') as cf_fd:
		cf_fd.write(conf_line)

	tmux_version = get_tmux_version(get_fallback_logger())

	dim = MutableDimensions(rows=50, cols=200)

	def prepare_test_1(p):
		sleep(5)

	def prepare_test_2(p):
		dim.cols = 40
		p.resize(dim)
		sleep(5)

	base_attrs = {
		((0, 0, 0), (243, 243, 243), 1, 0, 0): 'lead',
		((243, 243, 243), (11, 11, 11), 0, 0, 0): 'leadsep',
		((255, 255, 255), (11, 11, 11), 0, 0, 0): 'bg',
		((199, 199, 199), (88, 88, 88), 0, 0, 0): 'cwd',
		((88, 88, 88), (11, 11, 11), 0, 0, 0): 'cwdhsep',
		((0, 0, 0), (0, 224, 0), 0, 0, 0): 'defstl',
	}
	tests = (
		{
			'expected_result': get_expected_result(
				tmux_version,
				expected_result_old=(
					'{lead: 0 }{leadsep: }{bg: S2 string here  }'
					'{4: 0  }{cwdhsep:| }{6:bash  }'
					'{bg: }{4: 1- }{cwdhsep:| }{6:bash  }'
					'{bg: }{7: }{8:2* | }{9:bash }{10: }'
					'{bg:' + (' ' * 124) + '}'
					'{cwdhsep: }{cwd: S1 string here }', updated(base_attrs, {
						((133, 133, 133), (11, 11, 11), 0, 0, 0): 4,
						((188, 188, 188), (11, 11, 11), 0, 0, 0): 6,
						((11, 11, 11), (0, 102, 153), 0, 0, 0): 7,
						((102, 204, 255), (0, 102, 153), 0, 0, 0): 8,
						((255, 255, 255), (0, 102, 153), 1, 0, 0): 9,
						((0, 102, 153), (11, 11, 11), 0, 0, 0): 10,
					})),
				expected_result_1_8=(
					'{lead: 0 }{leadsep: }{bg: S2 string here  }'
					'{4: 0  }{cwdhsep:| }{6:bash  }'
					'{bg: }{4: 1- }{cwdhsep:| }{7:bash  }'
					'{bg: }{8: }{9:2* | }{10:bash }{7: }'
					'{bg:' + (' ' * 124) + '}'
					'{cwdhsep: }{cwd: S1 string here }', updated(base_attrs, {
						((133, 133, 133), (11, 11, 11), 0, 0, 0): 4,
						((188, 188, 188), (11, 11, 11), 0, 0, 0): 6,
						((0, 102, 153), (11, 11, 11), 0, 0, 0): 7,
						((11, 11, 11), (0, 102, 153), 0, 0, 0): 8,
						((102, 204, 255), (0, 102, 153), 0, 0, 0): 9,
						((255, 255, 255), (0, 102, 153), 1, 0, 0): 10,
					})),
				expected_result_2_0=(
					'{lead: 0 }{leadsep: }{bg: S2 string here }'
					'{4: 0  }{cwdhsep:| }{6:bash  }'
					'{bg: }{4: 1- }{cwdhsep:| }{7:bash  }'
					'{bg: }{8: }{9:2* | }{10:bash }{7: }'
					'{bg:' + (' ' * 125) + '}'
					'{cwdhsep: }{cwd: S1 string here }', updated(base_attrs, {
						((133, 133, 133), (11, 11, 11), 0, 0, 0): 4,
						((188, 188, 188), (11, 11, 11), 0, 0, 0): 6,
						((0, 102, 153), (11, 11, 11), 0, 0, 0): 7,
						((11, 11, 11), (0, 102, 153), 0, 0, 0): 8,
						((102, 204, 255), (0, 102, 153), 0, 0, 0): 9,
						((255, 255, 255), (0, 102, 153), 1, 0, 0): 10,
					})),
			),
			'prep_cb': prepare_test_1,
			'row': dim.rows - 1,
		}, {
			'expected_result': get_expected_result(
				tmux_version,
				expected_result_old=('{bg:' + (' ' * 40) + '}', base_attrs),
				expected_result_1_7=(
					'{lead: 0 }'
					'{leadsep: }{bg: <}{4:h  }{bg: }{5: }'
					'{6:2* | }{7:bash }{8: }{bg: }{cwdhsep: }'
					'{cwd: S1 string here }', updated(base_attrs, {
						((188, 188, 188), (11, 11, 11), 0, 0, 0): 4,
						((11, 11, 11), (0, 102, 153), 0, 0, 0): 5,
						((102, 204, 255), (0, 102, 153), 0, 0, 0): 6,
						((255, 255, 255), (0, 102, 153), 1, 0, 0): 7,
						((0, 102, 153), (11, 11, 11), 0, 0, 0): 8,
					})),
				expected_result_1_8=(
					'{lead: 0 }'
					'{leadsep: }{bg: <}{4:h  }{bg: }{5: }'
					'{6:2* | }{7:bash }{4: }{bg: }{cwdhsep: }'
					'{cwd: S1 string here }', updated(base_attrs, {
						((0, 102, 153), (11, 11, 11), 0, 0, 0): 4,
						((11, 11, 11), (0, 102, 153), 0, 0, 0): 5,
						((102, 204, 255), (0, 102, 153), 0, 0, 0): 6,
						((255, 255, 255), (0, 102, 153), 1, 0, 0): 7,
					})),
				expected_result_2_0=(
					'{lead: 0 }'
					'{leadsep: }{bg:<}{4:ash  }{bg: }{5: }'
					'{6:2* | }{7:bash }{4: }{cwdhsep: }'
					'{cwd: S1 string here }', updated(base_attrs, {
						((0, 102, 153), (11, 11, 11), 0, 0, 0): 4,
						((11, 11, 11), (0, 102, 153), 0, 0, 0): 5,
						((102, 204, 255), (0, 102, 153), 0, 0, 0): 6,
						((255, 255, 255), (0, 102, 153), 1, 0, 0): 7,
					})),
			),
			'prep_cb': prepare_test_2,
			'row': dim.rows - 1,
		}
	)

	args = [
		# Specify full path to tmux socket (testing tmux instance must not 
		# interfere with user one)
		'-S', socket_path,
		# Force 256-color mode
		'-2',
		# Request verbose logging just in case
		'-v',
		# Specify configuration file
		'-f', conf_file,
		# Run bash three times
		'new-session', 'bash --norc --noprofile -i', ';',
		'new-window', 'bash --norc --noprofile -i', ';',
		'new-window', 'bash --norc --noprofile -i', ';',
	]

	with PowerlineTestSuite('tmux') as suite:
		return do_terminal_tests(
			tests=tests,
			cmd=tmux_exe,
			dim=dim,
			args=args,
			env=env,
			cwd=TEST_ROOT,
			fin_cb=tmux_fin_cb,
			last_attempt_cb=print_tmux_logs,
			suite=suite,
		)


if __name__ == '__main__':
	if main():
		raise SystemExit(0)
	else:
		raise SystemExit(1)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

import sys
import json
import logging

from itertools import count

try:
	import vim
except ImportError:
	vim = object()

from powerline.bindings.vim import vim_get_func, vim_getvar, get_vim_encoding, python_to_vim
from powerline import Powerline, FailedUnicode, finish_common_config
from powerline.lib.dict import mergedicts
from powerline.lib.unicode import u


def _override_from(config, override_varname, key=None):
	try:
		overrides = vim_getvar(override_varname)
	except KeyError:
		return config
	if key is not None:
		try:
			overrides = overrides[key]
		except KeyError:
			return config
	mergedicts(config, overrides)
	return config


class VimVarHandler(logging.Handler, object):
	'''Vim-specific handler which emits messages to Vim global variables

	:param str varname:
		Variable where
	'''
	def __init__(self, varname):
		super(VimVarHandler, self).__init__()
		utf_varname = u(varname)
		self.vim_varname = utf_varname.encode('ascii')
		vim.command('unlet! g:' + utf_varname)
		vim.command('let g:' + utf_varname + ' = []')

	def emit(self, record):
		message = u(record.message)
		if record.exc_text:
			message += '\n' + u(record.exc_text)
		vim.eval(b'add(g:' + self.vim_varname + b', ' + python_to_vim(message) + b')')


class VimPowerline(Powerline):
	def init(self, pyeval='PowerlinePyeval', **kwargs):
		super(VimPowerline, self).init('vim', **kwargs)
		self.last_window_id = 1
		self.pyeval = pyeval
		self.construct_window_statusline = self.create_window_statusline_constructor()
		if all((hasattr(vim.current.window, attr) for attr in ('options', 'vars', 'number'))):
			self.win_idx = self.new_win_idx
		else:
			self.win_idx = self.old_win_idx
			self._vim_getwinvar = vim_get_func('getwinvar', 'bytes')
			self._vim_setwinvar = vim_get_func('setwinvar')

	if sys.version_info < (3,):
		def create_window_statusline_constructor(self):
			window_statusline = b'%!' + str(self.pyeval) + b'(\'powerline.statusline({0})\')'
			return window_statusline.format
	else:
		def create_window_statusline_constructor(self):
			startstr = b'%!' + self.pyeval.encode('ascii') + b'(\'powerline.statusline('
			endstr = b')\')'
			return lambda idx: (
				startstr + str(idx).encode('ascii') + endstr
			)

	create_window_statusline_constructor.__doc__ = (
		'''Create function which returns &l:stl value being given window index

		Created function must return :py:class:`bytes` instance because this is 
		what ``window.options['statusline']`` returns (``window`` is 
		:py:class:`vim.Window` instance).

		:return:
			Function with type ``int → bytes``.
		'''
	)

	default_log_stream = sys.stdout

	def add_local_theme(self, key, config):
		'''Add local themes at runtime (during vim session).

		:param str key:
			Matcher name (in format ``{matcher_module}.{module_attribute}`` or 
			``{module_attribute}`` if ``{matcher_module}`` is 
			``powerline.matchers.vim``). Function pointed by 
			``{module_attribute}`` should be hashable and accept a dictionary 
			with information about current buffer and return boolean value 
			indicating whether current window matched conditions. See also 
			:ref:`local_themes key description <config-ext-local_themes>`.

		:param dict config:
			:ref:`Theme <config-themes>` dictionary.

		:return:
			``True`` if theme was added successfully and ``False`` if theme with 
			the same matcher already exists.
		'''
		self.update_renderer()
		matcher = self.get_matcher(key)
		theme_config = {}
		for cfg_path in self.theme_levels:
			try:
				lvl_config = self.load_config(cfg_path, 'theme')
			except IOError:
				pass
			else:
				mergedicts(theme_config, lvl_config)
		mergedicts(theme_config, config)
		try:
			self.renderer.add_local_theme(matcher, {'config': theme_config})
		except KeyError:
			return False
		else:
			# Hack for local themes support: when reloading modules it is not 
			# guaranteed that .add_local_theme will be called once again, so 
			# this function arguments will be saved here for calling from 
			# .do_setup().
			self.setup_kwargs.setdefault('_local_themes', []).append((key, config))
			return True

	get_encoding = staticmethod(get_vim_encoding)

	def load_main_config(self):
		main_config = _override_from(super(VimPowerline, self).load_main_config(), 'powerline_config_overrides')
		try:
			use_var_handler = bool(int(vim_getvar('powerline_use_var_handler')))
		except KeyError:
			use_var_handler = False
		if use_var_handler:
			main_config.setdefault('common', {})
			main_config['common'] = finish_common_config(self.get_encoding(), main_config['common'])
			main_config['common']['log_file'].append(['powerline.vim.VimVarHandler', [['powerline_log_messages']]])
		return main_config

	def load_theme_config(self, name):
		return _override_from(
			super(VimPowerline, self).load_theme_config(name),
			'powerline_theme_overrides',
			name
		)

	def get_local_themes(self, local_themes):
		if not local_themes:
			return {}

		return dict((
			(matcher, {'config': self.load_theme_config(val)})
			for matcher, key, val in (
				(
					(None if k == '__tabline__' else self.get_matcher(k)),
					k,
					v
				)
				for k, v in local_themes.items()
			) if (
				matcher or
				key == '__tabline__'
			)
		))

	def get_matcher(self, match_name):
		match_module, separator, match_function = match_name.rpartition('.')
		if not separator:
			match_module = 'powerline.matchers.{0}'.format(self.ext)
			match_function = match_name
		return self.get_module_attr(match_module, match_function, prefix='matcher_generator')

	def get_config_paths(self):
		try:
			return vim_getvar('powerline_config_paths')
		except KeyError:
			return super(VimPowerline, self).get_config_paths()

	def do_setup(self, pyeval=None, pycmd=None, can_replace_pyeval=True, _local_themes=()):
		import __main__
		if not pyeval:
			pyeval = 'pyeval' if sys.version_info < (3,) else 'py3eval'
			can_replace_pyeval = True
		if not pycmd:
			pycmd = get_default_pycmd()

		set_pycmd(pycmd)

		# pyeval() and vim.bindeval were both introduced in one patch
		if (not hasattr(vim, 'bindeval') and can_replace_pyeval) or pyeval == 'PowerlinePyeval':
			vim.command(('''
				function! PowerlinePyeval(e)
					{pycmd} powerline.do_pyeval()
				endfunction
			''').format(pycmd=pycmd))
			pyeval = 'PowerlinePyeval'

		self.pyeval = pyeval
		self.construct_window_statusline = self.create_window_statusline_constructor()

		self.update_renderer()
		__main__.powerline = self

		try:
			if (
				bool(int(vim.eval('has(\'gui_running\') && argc() == 0')))
				and not vim.current.buffer.name
				and len(vim.windows) == 1
			):
				# Hack to show startup screen. Problems in GUI:
				# - Defining local value of &statusline option while computing
				#   global value purges startup screen.
				# - Defining highlight group while computing statusline purges
				#   startup screen.
				# This hack removes the “while computing statusline” part: both 
				# things are defined, but they are defined right now.
				#
				# The above condition disables this hack if no GUI is running, 
				# Vim did not open any files and there is only one window. 
				# Without GUI everything works, in other cases startup screen is 
				# not shown.
				self.new_window()
		except UnicodeDecodeError:
			# vim.current.buffer.name may raise UnicodeDecodeError when using 
			# Python-3*. Fortunately, this means that current buffer is not 
			# empty buffer, so the above condition should be False.
			pass

		# Cannot have this in one line due to weird newline handling (in :execute 
		# context newline is considered part of the command in just the same cases 
		# when bar is considered part of the command (unless defining function 
		# inside :execute)). vim.command is :execute equivalent regarding this case.
		vim.command('augroup Powerline')
		vim.command('	autocmd! ColorScheme * :{pycmd} powerline.reset_highlight()'.format(pycmd=pycmd))
		vim.command('	autocmd! VimLeavePre * :{pycmd} powerline.shutdown()'.format(pycmd=pycmd))
		vim.command('augroup END')

		# Hack for local themes support after reloading.
		for args in _local_themes:
			self.add_local_theme(*args)

	def reset_highlight(self):
		try:
			self.renderer.reset_highlight()
		except AttributeError:
			# Renderer object appears only after first `.render()` call. Thus if 
			# ColorScheme event happens before statusline is drawn for the first 
			# time AttributeError will be thrown for the self.renderer. It is 
			# fine to ignore it: no renderer == no colors to reset == no need to 
			# do anything.
			pass

	def new_win_idx(self, window_id):
		r = None

		for window in vim.windows:
			try:
				curwindow_id = window.vars['powerline_window_id']
				if r is not None and curwindow_id == window_id:
					raise KeyError
			except KeyError:
				curwindow_id = self.last_window_id
				self.last_window_id += 1
				window.vars['powerline_window_id'] = curwindow_id
			statusline = self.construct_window_statusline(curwindow_id)
			if window.options['statusline'] != statusline:
				window.options['statusline'] = statusline
			if curwindow_id == window_id if window_id else window is vim.current.window:
				r = (window, curwindow_id, window.number)
		return r

	def old_win_idx(self, window_id):
		r = None
		for winnr, window in zip(count(1), vim.windows):
			curwindow_id = self._vim_getwinvar(winnr, 'powerline_window_id')
			if curwindow_id and not (r is not None and curwindow_id == window_id):
				curwindow_id = int(curwindow_id)
			else:
				curwindow_id = self.last_window_id
				self.last_window_id += 1
				self._vim_setwinvar(winnr, 'powerline_window_id', curwindow_id)
			statusline = self.construct_window_statusline(curwindow_id)
			if self._vim_getwinvar(winnr, '&statusline') != statusline:
				self._vim_setwinvar(winnr, '&statusline', statusline)
			if curwindow_id == window_id if window_id else window is vim.current.window:
				r = (window, curwindow_id, winnr)
		return r

	def statusline(self, window_id):
		window, window_id, winnr = self.win_idx(window_id) or (None, None, None)
		if not window:
			return FailedUnicode('No window {0}'.format(window_id))
		return self.render(window, window_id, winnr)

	def tabline(self):
		r = self.win_idx(None)

		if r:
		    return self.render(*r, is_tabline=True)
		else:
		    win = vim.current.window
		    r = (
			win,
			win.vars.get('powerline_window_id', self.last_window_id),
			win.number,
		    )
		    return self.render(*r, is_tabline=True)

	def new_window(self):
		return self.render(*self.win_idx(None))

	@staticmethod
	def do_pyeval():
		'''Evaluate python string passed to PowerlinePyeval

		Is here to reduce the number of requirements to __main__ globals to just 
		one powerline object (previously it required as well vim and json).
		'''
		import __main__
		vim.command('return ' + json.dumps(eval(vim.eval('a:e'), __main__.__dict__)))

	def setup_components(self, components):
		if components is None:
			components = ('statusline', 'tabline')
		if 'statusline' in components:
			# Is immediately changed after new_window function is run. Good for 
			# global value.
			vim.command('set statusline=%!{pyeval}(\'powerline.new_window()\')'.format(
				pyeval=self.pyeval))
		if 'tabline' in components:
			vim.command('set tabline=%!{pyeval}(\'powerline.tabline()\')'.format(
				pyeval=self.pyeval))


pycmd = None


def set_pycmd(new_pycmd):
	global pycmd
	pycmd = new_pycmd


def get_default_pycmd():
	return 'python' if sys.version_info < (3,) else 'python3'


def setup(*args, **kwargs):
	powerline = VimPowerline()
	return powerline.setup(*args, **kwargs)



# vim:fileencoding=utf-8:noet
from __future__ import (unicode_literals, division, absolute_import, print_function)

from datetime import datetime


def date(pl, format='%Y-%m-%d', istime=False, timezone=None):
	'''Return the current date.

	:param str format:
		strftime-style date format string
	:param bool istime:
		If true then segment uses ``time`` highlight group.
	:param string timezone:
		Specify a timezone to use as ``+HHMM`` or ``-HHMM``.
		(Defaults to system defaults.)

	Divider highlight group used: ``time:divider``.

	Highlight groups used: ``time`` or ``date``.
	'''

	try:
		tz = datetime.strptime(timezone, '%z').tzinfo if timezone else None
	except ValueError:
		tz = None

	nw = datetime.now(tz)

	try:
		contents = nw.strftime(format)
	except UnicodeEncodeError:
		contents = nw.strftime(format.encode('utf-8')).decode('utf-8')

	return [{
		'contents': contents,
		'highlight_groups': (['time'] if istime else []) + ['date'],
		'divider_highlight_group': 'time:divider' if istime else None,
	}]


UNICODE_TEXT_TRANSLATION = {
	ord('\''): '’',
	ord('-'): '‐',
}


def fuzzy_time(pl, format='{minute_str} {hour_str}', unicode_text=False, timezone=None, hour_str=['twelve', 'one', 'two', 'three', 'four',
    'five', 'six', 'seven', 'eight', 'nine', 'ten', 'eleven'], minute_str = {
	'0':  'o\'clock', '5':  'five past', '10': 'ten past','15': 'quarter past',
	'20': 'twenty past', '25': 'twenty-five past', '30': 'half past', '35': 'twenty-five to',
	'40': 'twenty to', '45': 'quarter to', '50': 'ten to', '55': 'five to'
	}, special_case_str = {
	    '(23, 58)': 'round about midnight',
	    '(23, 59)': 'round about midnight',
	    '(0, 0)': 'midnight',
	    '(0, 1)': 'round about midnight',
	    '(0, 2)': 'round about midnight',
	    '(12, 0)': 'noon',
	}):

	'''Display the current time as fuzzy time, e.g. "quarter past six".

	:param string format:
		Format used to display the fuzzy time. (Ignored when a special time
		is displayed.)
	:param bool unicode_text:
		If true then hyphenminuses (regular ASCII ``-``) and single quotes are
		replaced with unicode dashes and apostrophes.
	:param string timezone:
		Specify a timezone to use as ``+HHMM`` or ``-HHMM``.
		(Defaults to system defaults.)
	:param string list hour_str:
		Strings to be used to display the hour, starting with midnight.
		(This list may contain 12 or 24 entries.)
	:param dict minute_str:
		Dictionary mapping minutes to strings to be used to display them.
	:param dict special_case_str:
		Special strings for special times.

	Highlight groups used: ``fuzzy_time``.
	'''

	try:
		tz = datetime.strptime(timezone, '%z').tzinfo if timezone else None
	except ValueError:
		tz = None

	now = datetime.now(tz)

	try:
		# We don't want to enforce a special type of spaces/ alignment in the input
		from ast import literal_eval
		special_case_str = {literal_eval(x):special_case_str[x] for x in special_case_str}
		result = special_case_str[(now.hour, now.minute)]
		if unicode_text:
			result = result.translate(UNICODE_TEXT_TRANSLATION)
		return result
	except KeyError:
		pass

	hour = now.hour
	if now.minute >= 30:
		hour = hour + 1
	hour = hour % len(hour_str)

	min_dis = 100
	min_pos = 0

	for mn in minute_str:
		mn = int(mn)
		if now.minute >= mn and now.minute - mn < min_dis:
			min_dis = now.minute - mn
			min_pos = mn
		elif now.minute < mn and mn - now.minute < min_dis:
			min_dis = mn - now.minute
			min_pos = mn
	result = format.format(minute_str=minute_str[str(min_pos)], hour_str=hour_str[hour])

	if unicode_text:
		result = result.translate(UNICODE_TEXT_TRANSLATION)

	return result
